# flake8: noqa

from .filtering import Stride2D, TargetStride, UndoStride2D
from .modules import Argmax, Lambda, Sign, SignFG
from .reversibles import (
    BatchNorm1DReversible,
    LeakyReLUInvertable,
    Neg1ToZero,
    Null,
    Reversible,
    SequenceReversible,
    SigmoidInvertable,
    TargetReverser,
    ZeroToNeg1,
)
from .utils import (
    add_prev,
    binary_encoding,
    calc_correlation_mae,
    calc_size2d,
    calc_stride2d,
    chain,
    coalesce,
    create_dataloader,
    detach,
    expand_dim0,
    freshen,
    get_indexed,
    get_model_parameters,
    repeat_on_indices,
    sequential,
    set_parameters,
    to_float,
    to_np,
    to_signed_neg,
    to_th,
    to_th_as,
    to_zero_neg,
    update,
    update_model_parameters,
    update_param,
)



# 1st party
import typing

# 3rd party
import torch
import torch.nn as nn
from torch.nn.parameter import Parameter

from ..kaku import IO, Assessment
from ..kaku import Reduction

# local
from ..utils import get_model_parameters, update_model_parameters


def cat_params(
    params: torch.Tensor, perturbed_params: torch.Tensor, reorder: bool = False
):
    """Reorder the parameters for the perturber

    Args:
        value (torch.Tensor): _description_
        perturbed (torch.Tensor): _description_

    Returns:
        _type_: _description_
    """
    if params.shape != perturbed_params.shape[1:]:
        raise RuntimeError(
            f"The parameters shape {params.shape} does not match "
            f"the perturbed_params shape {perturbed_params.shape}"
        )
    ordered = torch.cat([params[None], perturbed_params])
    if reorder:
        reordered = torch.randperm(len(perturbed_params) + 1, device=params.device)
        return ordered[reordered]
    return ordered


def expand(x: torch.Tensor, k: int):
    """Expand an input to repeat k times"""
    return x[None].repeat(k, *([1] * len(x.shape)))


def flatten(x: torch.Tensor):
    """Flatten the population and batch dimensions of a population"""
    return x.view(x.shape[0] * x.shape[1], *x.shape[2:])


def deflatten(x: torch.Tensor, k: int) -> torch.Tensor:
    """Deflatten the population and batch dimensions of a population"""
    if x.dim() == 0:
        raise ValueError("Input dimension == 0")

    return x.view(k, -1, *x.shape[1:])


def binary_prob(
    x: torch.Tensor, loss: torch.Tensor, retrieve_counts: bool = False
) -> typing.Union[torch.Tensor, typing.Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]:
    """

    Args:
        x (torch.Tensor): The population input
        loss (torch.Tensor): The loss
        retrieve_counts (bool, optional): Whether to return the positive 
          and negative counts in the result. Defaults to False.

    Returns:
        typing.Union[ torch.Tensor, typing.Tuple[torch.Tensor, torch.Tensor, torch.Tensor] ]: _description_
    """
    is_pos = (x == 1).unsqueeze(-1)
    is_neg = ~is_pos
    pos_count = is_pos.sum(dim=0)
    neg_count = is_neg.sum(dim=0)
    positive_loss = (loss[:, :, None] * is_pos.float()).sum(dim=0) / pos_count
    negative_loss = (loss[:, :, None] * is_neg.float()).sum(dim=0) / neg_count
    updated = (positive_loss < negative_loss).type_as(x).mean(dim=-1)

    if not retrieve_counts:
        return updated

    return updated, pos_count.squeeze(-1), neg_count.squeeze(-1)


def gaussian_sample(
    mean: torch.Tensor, std: torch.Tensor, k: int = None
) -> torch.Tensor:
    """generate a sample from a gaussian

    Args:
        mean (torch.Tensor): _description_
        std (torch.Tensor): _description_
        k (int): The number of samples to generate. If None will generate 1 sample and the dimension
         will not be expanded

    Returns:
        torch.Tensor: The sample or samples generated
    """
    if k is not None:
        if k <= 0:
            raise ValueError(f"Argument {k} must be greater than 0")
        return (
            torch.randn([k, *mean.shape], device=mean.device, dtype=mean.dtype)
            * std[None]
            + mean[None]
        )
    return torch.randn_like(mean) * std + mean


def gather_idx_from_population(pop: torch.Tensor, idx: torch.LongTensor):
    """Retrieve the indices from population. idx is a 2 dimensional tensor"""
    repeat_by = [1] * len(idx.shape)
    for i, sz in enumerate(pop.shape[2:]):
        idx = idx.unsqueeze(i + 2)
        repeat_by.append(sz)
    idx = idx.repeat(*repeat_by)
    return pop.gather(0, idx)


def select_best_individual(
    pop_val: torch.Tensor, assessment: Assessment
) -> torch.Tensor:
    """
    Args:
        pop_val (torch.Tensor): The tensor for the population
        assessment (Assessment): The evaluation of the individuals in a population

    Returns:
        Tensor: the best individual in the population
    """

    _, idx = assessment.best(0, True)
    return pop_val[idx[0]]


def select_best_feature(pop_val: torch.Tensor, assessment: Assessment) -> torch.Tensor:
    """
    Args:
        pop_val (torch.Tensor): The population to select from
        assessment (Assessment): The evaluation of the features in the population

    Returns:
        torch.Tensor: The best features in the population
    """

    value = assessment.value
    if assessment.maximize:
        idx = value.argmax(0, True)
    else:
        idx = value.argmin(0, True)

    pop_val = pop_val.view(value.shape[0], value.shape[1], -1)
    idx = idx[:, :, None].repeat(1, 1, pop_val.shape[2])
    return pop_val.gather(0, idx).squeeze(0)


class Individual(object):
    """An individual in a population"""

    def __init__(
        self,
        assessment: Assessment = None,
        **values: typing.Union[nn.Module, torch.Tensor, Parameter],
    ):
        """initializer"""

        self._assessment = assessment
        self._population = None
        self._id = None
        self._parameters = {}

        for k, v in values.items():
            if isinstance(v, nn.Module):
                v = get_model_parameters(v)
            self._parameters[k] = v
        # TODO: validate the sizes

    def set_model(self, model: nn.Module, key: str) -> "Individual":
        """Update the parameters in a module

        Args:
            model (nn.Module): The model to update
            key (str): The key to the values in the Individual to update the parameter with

        Returns:
            Individual: self
        """
        update_model_parameters(model, self._parameters[key])
        return self

    def set_p(self, parameter: Parameter, key: str) -> "Individual":
        """Set a nn.parameter.Parameter variable with values in the individual

        Args:
            parameter (Parameter): Parameters to set
            key (str): The key to the values in the Individual to update the parameter with

        Returns:
            Individual: self
        """
        parameter.data = self._parameters[key]
        return self

    def join(self, population: "Population", individual_idx: int) -> "Individual":
        """Set the population for the individual

        Args:
            population (Population): The population it joins
            individual_idx (int): The index for the individual

        Returns:
            Individual: self
        """
        if not population.authenticate(self, individual_idx):
            raise ValueError("Cannot authenticate individual in the population")
        self._population = population
        self._id = individual_idx
        return self

    def report(self, assessment: Assessment) -> "Individual":
        """Report the assessment for an individual. If the individual in a population
        it will set the assessment in the population as well

        Args:
            assessment (Assessment): The assessment for the individual

        Returns:
            Individual: self
        """
        self._assessment = assessment
        if self._population is not None:
            self._population.report_for(self._id, assessment)
        return self

    def __iter__(self) -> typing.Iterator:
        """Iterate over the values in the individual

        Yields:
            Iterator[typing.Iterator]: The iterator to iterate over the values with
        """

        for k, v in self._parameters.items():
            yield k, v

    def __getitem__(self, key: str) -> typing.Union[torch.Tensor, Parameter]:
        """Retrieve the value specified by key
        Args:
            key (str):

        Returns:
            typing.Union[torch.Tensor, Parameter]: The value in the key
        """
        return self._parameters[key]

    def __contains__(self, key: str) -> bool:
        """
        Args:
            key (str): The key to the values in the Individual to update the parameter with
        Returns:
            bool: True if the individual contains the key
        """
        return key in self._parameters

    @property
    def assessment(self) -> Assessment:
        return self._assessment


class Population(object):
    """Collection of individuals"""

    def __init__(self, **kwargs: typing.Union[torch.Tensor, Parameter]):
        """initializer

        name<str>: Value<Tensor>

        Raises:
            ValueError: If dimension is 0 for any
            ValueError: If the population size is not the same as all
        """
        self._parameters = {}
        k = None
        for key, v in kwargs.items():
            if v.dim() == 0:
                raise ValueError("Population must consist of tensors of dimension > 0")
            if k is not None and v.shape[0] != k:
                raise ValueError(
                    "All members of the population must have the same size"
                )
            k = k or v.shape[0]
            self._parameters[key] = v
        self._k = k
        # lazily fill this in if requested
        self._individuals = {}
        self._assessments: typing.List[Assessment] = [None] * self._k
        self._assessment_size = None

    @property
    def k(self) -> int:
        """
        Returns:
            int: Number of members in the population
        """
        return self._k

    def authenticate(self, individual: Individual, index: int) -> bool:
        """
        Args:
            individual (Individual): The individual to check
            index (int): the index for the individual

        Returns:
            bool: Whether the individual is in the population
        """

        return self._individuals.get(index) is not individual

    def get_assessment(self, i: int) -> Assessment:
        """
        Args:
            i (int): index of individual

        Returns:
            Assessment: Assessment for an individual
        """
        if self._assessments is None:
            return None
        return self._assessments[i]

    def get_i(self, i: int) -> Individual:
        """Retrieve an individual and their assessment

        Args:
            i (int): index of the individual

        Returns:
            Individual:
        """
        if i in self._individuals:
            return self._individuals[i]

        individual = Individual(
            **{f: self._parameters[f][i] for f in self._parameters.keys()},
            assessment=self.get_assessment(i),
        )
        individual.join(self, i)
        return individual

    def report(self, assessment: "Assessment") -> "Population":
        """Report the result of an assessment

        Args:
            assessment (Assessment): The assessment for the population

        Raises:
            ValueError: if the assessment is not the same size as the population

        Returns:
            Population: self
        """
        if len(assessment) != self._k:
            raise ValueError(
                "Length of assessment must be same "
                f"as population {self._k} not {len(assessment)}"
            )
        self._assessments = list(assessment)
        self._assessment_size = assessment.value.shape[1:]
        return self

    def report_for(self, id: int, assessment: "Assessment"):
        if id < 0 or id >= self._k:
            raise ValueError(f"Value i must be in range [0, {self._k})")
        if (
            self._assessment_size is not None
            and assessment.value.size() != self._assessment_size
        ):
            raise ValueError(
                f"Assessment size must be the same as others {self._assessment_size}"
            )
        self._assessment_size = self._assessment_size or assessment.value.size()
        self._assessments[id] = assessment

    def set_model(self, model: nn.Module, key: str, id: int):
        update_model_parameters(model, self._parameters[key][id])
        return self

    def set_p(
        self, parameter: Parameter, key: str, individual_index: int
    ) -> "Individual":
        """_summary_

        Args:
            parameter (Parameter): The parameter value to update
            key (str): The key to the values in the Individual to update the parameter with
            individual_index (int): The index of the individual

        Returns:
            Individual: self
        """
        parameter.data = self._parameters[key][individual_index]
        return self

    def individuals(self) -> typing.Iterator[Individual]:
        """
        Yields:
            Iterator[typing.Iterator[Individual]]: The individuals in the population
        """
        for i in range(self.k):
            yield self.get_i(i)

    def __len__(self) -> int:
        """
        Returns:
            int: The number of individuals in the population
        """
        return self._k

    @property
    def assessments(self) -> typing.List[Assessment]:
        """The assessments for the population

        Returns:
            typing.List[Assessment]:
        """
        return self._assessments

    def assessments_reported(self) -> bool:
        for assessment in self._assessments:
            if assessment is None:
                return False
        return True

    def stack_assessments(self) -> Assessment:
        """Stack all of the assessments

        Raises:
            ValueError: If an individual in the population has not been assessed

        Returns:
            Assessment: The assessments for the population
        """
        values = []

        for i, assessment in enumerate(self._assessments):
            if assessment is None:
                raise ValueError(f"Assessment {i} has not been set.")
            values.append(assessment.value)
        return Assessment(torch.stack(values), self._assessments[0].maximize)

    def __contains__(self, key: str) -> bool:
        """
        Args:
            key (str): The key to check if the individual contains

        Returns:
            bool: If the key is in the parameters
        """
        return key in self._parameters

    def _flattened_helper(self, key: str) -> torch.Tensor:
        """

        Args:
            key (str): The key for the value

        Returns:
            torch.Tensor:
        """
        t = self._parameters[key]
        return t.reshape(t.size(0) * t.size(1), *t.shape[2:])

    def flattened(
        self, key: typing.Union[str, typing.List[str]]
    ) -> typing.Union[torch.Tensor, typing.List[torch.Tensor]]:
        """Returned a flattend value

        Args:
            key (typing.Union[str, typing.List[str]]): The key for the value

        Returns:
            typing.Union[torch.Tensor, typing.List[torch.Tensor]]: Check if the
        """
        if isinstance(key, str):
            return self._flattened_helper(key)
        ts = []
        for key_i in key:
            ts.append(self._flattened_helper(key_i))
        return ts

    def __getitem__(
        self, key: typing.Union[str, typing.Tuple[str, int]]
    ) -> torch.Tensor:
        """
        Args:
            key (Union[str, typing.Tuple[str, int]]): _description_

        Returns:
            torch.Tensor: The value at key
        """
        if isinstance(key, tuple):
            field, i = key
            return self._parameters[field][i]
        return self._parameters[key]

    def __iter__(self) -> typing.Iterator[torch.Tensor]:
        """

        Yields:
            Iterator[typing.Iterator[torch.Tensor]]: Iterate over the values in the population
        """
        for k, v in self._parameters.items():
            yield k, v


def reduce_assessment_dim0(
    assessment: Assessment, k: int, reduction: str = "mean"
) -> Assessment:
    """
    Args:
        assessment (Assessment): The assessment for the population
        k (int): The size of the population
        reduction (str, optional): The name of the reduction. Defaults to "mean".

    Returns:
        Assessment: The reduced assessment
    """
    return Assessment(
        Reduction.sample_reduce_by(assessment.value.view(k, -1).value, reduction)
    )


def reduce_assessment_dim1(
    assessment: Assessment, k: int, flattened: bool = True, reduction: str = "mean"
) -> Assessment:
    """
    Args:
        assessment (Assessment): The assessment for the population
        k (int): The size of the population
        flattened (bool, optional): Whether the . Defaults to True.
        reduction (str, optional): The name of the reduction.. Defaults to "mean".

    Returns:
        Assessment: The reduced assessment
    """

    if not flattened:
        value = assessment.value.view(k * assessment.value.size(1))
    else:
        value = assessment.value

    return Assessment(Reduction.sample_reduce_by(value, reduction).view(k, -1))


def expand_t(t: IO, k: int) -> IO:
    """expand the population dimension for t

    Args:
        t (IO): the target IO
        k (int): the size of the population

    Returns:
        IO: the expanded target IO
    """

    ts = []
    for t_i in t:
        ts.append(flatten(expand(t_i, k)))

    return IO(*ts)


# TO DECIDE
# class Updater(ABC):

#     @abstractmethod
#     def __call__(self, population: Population) -> Population:
#         pass


# def expand_dim0(x: torch.Tensor, k: int, reshape: bool=True):
#     y = x[None].repeat(k, *([1] * len(x.size()))) #.transpose(0, 1)
#     if reshape:
#         return y.view(y.shape[0] * y.shape[1], *y.shape[2:])
#     return y



# 1st Party
import math
import typing
from functools import singledispatch

import numpy as np
import torch
import torch.nn as nn
import torch.nn.utils as nn_utils

# 3rd party
from torch.nn.utils import parameters_to_vector, vector_to_parameters
from torch.utils import data as torch_data


def to_np(x: torch.Tensor) -> np.ndarray:
    return x.detach().cpu().numpy()


def to_th(
    x: np.ndarray,
    dtype: torch.dtype,
    device: torch.device,
    requires_grad: bool = False,
    retains_grad: bool = False,
) -> torch.Tensor:
    """

    Args:
        x (np.ndarray): Array to convert
        dtype (torch.dtype): type to convert to
        device (torch.device): device to convert to
        requires_grad (bool, optional): Whether the tensor requires grad. Defaults to False.
        retains_grad (bool, optional): Whether the tensor retains a grad. Defaults to False.

    Returns:
        torch.Tensor: _description_
    """
    x: torch.Tensor = torch.tensor(
        x, dtype=dtype, requires_grad=requires_grad, device=device
    )
    if retains_grad:
        x.retain_grad()
    return x


def to_th_as(
    x: np.ndarray,
    as_: torch.Tensor,
    requires_grad: bool = False,
    retains_grad: bool = False,
) -> torch.Tensor:
    """

    Args:
        x (np.ndarray): _description_
        as_ (torch.Tensor): _description_
        requires_grad (bool, optional): _description_. Defaults to False.
        retains_grad (bool, optional): _description_. Defaults to False.

    Returns:
        torch.Tensor: _description_
    """

    x: torch.Tensor = torch.tensor(
        x, dtype=as_.dtype, requires_grad=requires_grad, device=as_.device
    )
    if retains_grad:
        x.retain_grad()
    return x


def chain(x, torch_modules: typing.Iterable[nn.Module]):

    for module in torch_modules:
        x = module.forward(x)
    return x


def coalesce(value, default_value) -> typing.Any:
    """
    Args:
        value: Value to coalesce if none
        default_value: Value to coalesce to

    Returns:
        Any: Either the default value or the value if not None
    """

    return value if value is not None else default_value


def batch_flatten(x: torch.Tensor):
    return x.view(x.size(0), -1)


def expand_dim0(x: torch.Tensor, k: int, reshape: bool = True):
    y = x[None].repeat(k, *([1] * len(x.size())))  # .transpose(0, 1)
    if reshape:
        return y.view(y.shape[0] * y.shape[1], *y.shape[2:])
    return y


def freshen(x: torch.Tensor, requires_grad: bool = True, inplace: bool = False):
    if not isinstance(x, torch.Tensor):
        return x
    if inplace:
        x.detach_()
    else:
        x = x.detach()
    x = x.requires_grad_(requires_grad)
    x.retain_grad()
    return x


def set_parameters(parameters: torch.Tensor, net: nn.Module):
    vector_to_parameters(parameters, net.parameters())


def get_parameters(net: nn.Module):
    return parameters_to_vector(net.parameters())


def to_float(x: typing.List[torch.Tensor]):
    return list(map(lambda xi: xi.mean().item(), x))


def update(
    current: torch.Tensor, new_: torch.Tensor, lr: typing.Optional[float] = None
):
    assert lr is None or (0.0 <= lr <= 1.0)
    if lr is not None:
        new_ = (lr * new_) + (1 - lr) * (current)
    return new_


def update_param(
    current: torch.Tensor, new_: torch.Tensor, lr: typing.Optional[float] = None
):
    p = nn.parameter.Parameter(update(current, new_, lr).detach())
    return p


def to_zero_neg(x: torch.Tensor) -> torch.Tensor:
    """convert a 'signed' binary tensor to have zeros for negatives

    Args:
        x (torch.Tensor): Signed binary tensor

    Returns:
        torch.Tensor: The binary tensor with negatives as zero
    """

    return (x + 1) / 2


def to_signed_neg(x: torch.Tensor) -> torch.Tensor:
    """convert a 'zero' binary tensor to have negative ones for negatives

    Args:
        x (torch.Tensor): Binary tensor with zeros for negatives

    Returns:
        torch.Tensor: The signed binary tensor
    """
    return (x * 2) - 1


def add_prev(cur, prev=None):

    if cur is not None and prev is not None:
        return cur + prev
    if cur is not None:
        return cur

    return prev


def create_dataloader(
    x: torch.Tensor,
    t: torch.Tensor,
    batch_size: int = 64,
    shuffle: bool = True,
    get_indices: bool = False,
):
    """Create data loader to loop over an input

    Args:
        x (torch.Tensor): _description_
        t (torch.Tensor): _description_
        batch_size (int, optional): _description_. Defaults to 64.
        shuffle (bool, optional): _description_. Defaults to True.
        get_indices (bool, optional): _description_. Defaults to False.

    Returns:
        _type_: _description_
    """
    if get_indices:
        indices = torch.range(0, len(x))
        dataset = torch_data.TensorDataset(x, t, indices)
    else:
        dataset = torch_data.TensorDataset(x, t)
    return torch_data.DataLoader(dataset, batch_size, shuffle)


def get_indexed(
    x: torch.Tensor, indices: typing.Optional[torch.LongTensor] = None
) -> torch.Tensor:

    if indices is not None:
        return x[indices]

    return x


def repeat_on_indices(
    x: torch.Tensor, t: torch.Tensor, indices: torch.LongTensor, iterations: int
):
    """Convenience function to loop over indieces

    Args:
        x (torch.Tensor): The input tensor
        t (torch.Tensor): The target tensor
        indices (torch.LongTensor): The indices to retrieve
        iterations (int): Number of times to iterate

    Yields:
        torch.Tensor, torch.Tensor : The sampled input and target tensor
    """

    x = get_indexed(x, indices)
    t = get_indexed(t, indices)

    for i in range(iterations):

        yield x, t


def get_model_parameters(model: nn.Module) -> torch.Tensor:
    """Convenience function to retrieve the parameters of a model

    Args:
        model (nn.Module): _description_

    Returns:
        torch.Tensor: _description_
    """

    return nn_utils.parameters_to_vector(model.parameters())


def update_model_parameters(model: nn.Module, theta: torch.Tensor):
    """Convenience function to update the parameters of a model

    Args:
        model (nn.Module): Model to update parameters for
        theta (torch.Tensor): The new parameters for the model
    """

    nn_utils.vector_to_parameters(theta, model.parameters())


def calc_correlation_mae(x1: torch.Tensor, x2: torch.Tensor):
    """Calculate the mean absolute error in correlation

    Args:
        x1 (torch.Tensor):
        x2 (torch.Tensor): _description_

    Returns:
        _type_: _description_
    """

    corr1 = torch.corrcoef(batch_flatten(x1))
    corr2 = torch.corrcoef(batch_flatten(x2))
    return torch.abs(corr1 - corr2).mean()


def calc_stride2d(x: torch.Tensor, stride):
    return (x.stride(0), x.stride(1), x.stride(2), stride[0], x.stride(2), stride[1])


def calc_size2d(x: torch.Tensor, stride, kernel_size):
    return (
        x.size(0),
        x.size(1),
        (x.size(2) - (kernel_size[0] - 1)) // stride[0],
        (x.size(3) - (kernel_size[1] - 1)) // stride[1],
        kernel_size[0],
        kernel_size[1],
    )


@singledispatch
def to_2dtuple(value: int) -> typing.Tuple[int, int]:
    return (value, value)


@to_2dtuple.register
def _(value: tuple) -> typing.Tuple[int, int]:
    return value


def detach(x: typing.Union[typing.Iterable[torch.Tensor], torch.Tensor]):

    if isinstance(x, list) or isinstance(x, tuple):
        result = []
        for x_i in x:
            if isinstance(x_i, torch.Tensor):
                result.append(x_i.detach())
            else:
                result.append(x_i)
        return result

    return x.detach() if isinstance(x, torch.Tensor) else x


def binary_encoding(
    x: torch.LongTensor, n_size: int, bit_size: bool = False
) -> torch.Tensor:

    if not bit_size:
        n_size = int(math.ceil(math.log2(n_size)))
    results = []
    for _ in range(n_size):
        results.append(x)
        x = x >> 1
    results = torch.stack(tuple(reversed(results))) & 1
    shape = list(range(results.dim()))
    shape = shape[1:] + shape[0:1]
    return results.permute(*shape)


def sequential(modules: typing.List[nn.Module], x: torch.Tensor) -> torch.Tensor:

    for module in modules:
        x = module(x)
    return x


# class DatasetLoader(ABC):

#     @abstractmethod
#     def training(self) -> torch_data.Dataset:
#         pass

#     @abstractmethod
#     def testing(self) -> torch_data.Dataset:
#         pass

#     @abstractmethod
#     def validation(self) -> torch_data.Dataset:
#         pass



# 1st party
import typing
from abc import ABC, abstractmethod

# 3rd party
from ..kaku import AssessmentDict, Learner


class Material(ABC):
    """Class used for retrieving training data or testing data"""

    @abstractmethod
    def __iter__(self) -> typing.Iterator:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass


class Classroom:
    """Stores the students in a class."""

    def __init__(self, **students: Learner):
        """initializer"""
        super().__init__()
        self._students = students

    def __getitem__(self, key: str) -> Learner:
        """retrieve the student

        Args:
            key (str): the name of the student

        Returns:
            Learner: The studennt
        """
        return self._students[key]

    def __setitem__(self, key: str, student: Learner):
        """Add a student to the classroom

        Args:
            key (str): the name of the student
            student (Learner): the instance for the student
        """

        self._students[key] = student

    def get(self, student: typing.Union[str, Learner, None]) -> Learner:
        """Convenience method to retrieve a student from the classroom.

        Args:
            learner (typing.Union[str, Learner, None]): Learner to retrieve. If it is an
            instance of Learner it will just return that student
        Returns:
            Learner: The learner
        """
        if isinstance(student, Learner):
            return student
        if student is not None:
            return self._students[student]
        return None


class Desk(object):
    def __init__(
        self,
        materials: typing.Dict[str, Material] = None,
        info: typing.Dict[str, typing.Any] = None,
    ):
        """Stores information for sharing between teachers and the materials

        Args:
            materials (typing.Dict[str, Material], optional): The materials used in the class. Defaults to None.
            info (typing.Dict[str, typing.Any], optional): Generic data to 
              share info between teachers related to teaching. Defaults to None.
        """
        super().__init__()
        self._materials = materials or {}
        self._info = info or {}

    def get_material(self, material: typing.Union[str, Material, None]) -> Material:
        """
        Args:
            material (typing.Union[str, Material, None]): _description_

        Returns:
            Material: the retrieved material
        """

        if isinstance(material, Material):
            return material
        if material is not None:
            return self._materials[material]
        return None

    def add_material(self, name: str, material: Material):
        """add a material to the desk

        Args:
            name (str): name of the material
            material (Material): the material
        """

        self._materials[name] = material

    def remove_material(self, name: str):
        """Remove a material from the desk

        Args:
            name (str): name of the material
        """

        del self._materials[name]

    def get_info(self, key: str) -> typing.Any:
        """Retrieve info from the desk

        Args:
            key (str): The

        Returns:
            typing.Any: _description_
        """
        return self._info[key]

    def add_info(self, key: str, info: typing.Any):
        """
        Args:
            key (str): the key for the info
            info (typing.Any): the info
        """
        self._info[key] = info

    def remove_info(self, key: str):
        """
        Args:
            key (str): the key for the info to remvoe
        """
        del self._info[key]


class Assistant(ABC):
    def __init__(self, name: str):

        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @abstractmethod
    def assist(
        self,
        teacher_name: str,
        assessment_dict: AssessmentDict = None,
        data: typing.Any = None,
    ):
        pass

    def __call__(
        self,
        teacher_name: str,
        assessment_dict: AssessmentDict = None,
        data: typing.Any = None,
    ):
        return self.assist(teacher_name, assessment_dict, data)


class AssistantTeam(object):
    def __init__(self, *assistants: Assistant):

        self._assistants: typing.Dict[str, Assistant] = {
            assistant.name: Assistant for assistant in assistants
        }

    def add(self, assistant: Assistant, overwrite: bool = True):
        """
        Args:
            assistant (Assistant): Add an assistant
            overwrite (bool, optional): _description_. Defaults to True.

        Raises:
            ValueError: _description_
        """

        if assistant.name in self._assistants:
            if not overwrite:
                raise ValueError()
            del self._assistants[assistant.name]
        self._assistants[assistant.name] = assistant

    def assist(
        self,
        teacher_name: str,
        assessment_dict: AssessmentDict = None,
        data: typing.Any = None,
    ):
        """Call all of the assistants in the team

        Args:
            teacher_name (str): The name of the teacher being assisted
            assessment_dict (AssessmentDict): The teachers assessment
            data (typing.Any): _description_
        """
        for assistant in self._assistants.values():
            assistant(teacher_name, assessment_dict, data)


class Teacher(ABC):
    def __init__(self, name: str):
        self._assistants = AssistantTeam()
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @abstractmethod
    def teach(self):
        pass

    def register(self, assistant: Assistant):
        if not hasattr(self, "_assistants"):
            self._assistants = {}
        self._assistants[assistant.name] = assistant

    def deregister(self, assistant: str):
        del self._assistants[assistant]

    def __call__(self, *args, **kwargs):
        self.teach(*args, **kwargs)


# class Assistant(ABC):

#     def __init__(self, name: str):
#         """An assistant teacher

#         Args:
#             name (str): The name of the teacher
#         """
#         self._name = name

#     @property
#     def name(self) -> str:
#         return self._name

#     @abstractmethod
#     def assist(self, teacher_name: str):
#         pass

#     def __call__(self, teacher_name: str):
#         return self.assist(teacher_name)



# 1st party
import typing

import torch

# 3rd Party
import torch.nn as nn

from ..kaku import AssessmentDict, OptimFactory, ThLoss

# Local
from ..kaku import (
    IO,
    BatchIdxStepTheta,
    BatchIdxStepX,
    Conn,
    Idx,
    LearningMachine,
    State,
    StepTheta,
    StepX,
    idx_io,
)


class GradStepTheta(StepTheta):
    """Update theta with the loss between y and t on the forward pass"""

    Y_NAME = "y"

    def __init__(
        self,
        learner: LearningMachine,
        optim_factory: OptimFactory,
        reduction: str = "mean",
    ):
        super().__init__()
        self.learner = learner
        self.optim = optim_factory(learner.parameters())
        self.reduction = reduction

    def step(self, conn: Conn, state: State, from_: IO = None) -> Conn:
        x, t, y = conn.step
        y = state.get(self, self.Y_NAME)
        stepped = state.get(self, "stepped", False)
        if stepped or y is None:
            y = self.learner(x, state, detach=False)

        self.optim.zero_grad()
        assessment = self.learner.assess_y(y, t)
        assessment.backward("loss")
        state.store(self, "stepped", True)
        self.optim.step()
        return conn.connect_in(from_)


class NullStepTheta(StepTheta):
    """Do not update theta"""

    def step(self, conn: Conn, state: State, from_: IO = None) -> Conn:
        return conn.connect_in(from_in_x=from_)


class GradLoopStepTheta(BatchIdxStepTheta):
    """Update theta with the loss between y and t after passing forward again"""

    def __init__(
        self,
        learner: LearningMachine,
        optim_factory: OptimFactory,
        reduction: str = "mean",
        loss_name: str = "loss",
    ):
        super().__init__()
        self.learner = learner
        self.optim = optim_factory(learner.parameters())
        self.reduction = reduction
        self.loss_name = loss_name

    def step(
        self, conn: Conn, state: State, from_: IO = None, batch_idx: Idx = None
    ) -> Conn:
        x = idx_io(conn.step.x, batch_idx, False)
        t = idx_io(conn.step.t, batch_idx, False)

        y = self.learner(x, state, False)

        self.optim.zero_grad()
        assessment = self.learner.assess_y(y, t, self.reduction)
        assessment[self.loss_name].backward()
        self.optim.step()

        return conn.connect_in(from_)


class GradStepX(StepX):
    """Update x with the loss between y and t based on the grad value of step_x.x"""

    Y_NAME = "y"

    def step_x(self, conn: Conn, state: State) -> Conn:

        x = conn.step_x.x[0]
        x = x - x.grad
        x.grad = None

        # TODO: Debug. This is causing problems in backpropagation
        # due to the inplace operation
        # update_io(IO(x), conn.step_x.x)
        conn.step_x.x = IO(x, detach=True)
        conn = conn.tie_step(True)
        return conn


class GradLoopStepX(BatchIdxStepX):
    """Update x with the loss between y and t after passing x forward again and getting the grad of x"""

    def __init__(
        self,
        learner: LearningMachine,
        optim_factory: OptimFactory,
        reduction: str = "mean",
        loss_name: str = "loss",
    ):
        """initializer

        Args:
            learner (LearningMachine):
            optim_factory (OptimFactory): OptimFactory for "optimizing" x
            reduction (str, optional): The loss reduction to use. Defaults to 'mean'.
            loss_name (str, optional): Name of the loss. Defaults to 'loss'.
        """
        super().__init__()
        self.learner = learner
        self.optim_factory = optim_factory
        self.reduction = reduction
        self.loss_name = loss_name

    def step_x(self, conn: Conn, state: State, batch_idx: Idx = None) -> Conn:
        my_state = conn.state.mine(self)
        if "optim" not in my_state:
            my_state.optim = self.optim_factory([*conn.step_x.x])
        x = idx_io(conn.step_x.x, batch_idx)
        t = idx_io(conn.step_x.t, batch_idx)
        my_state.optim.zero_grad()
        y = self.learner(x, detach=False)
        assessment = self.learner.assess_y(y, t, self.reduction)
        assessment.backward(self.loss_name)
        my_state.optim.step()

        # TODO: Detach
        return conn.tie_inp()


class GradLearner(LearningMachine):
    """Standard gradient learner"""

    VALIDATION_NAME = "validation"
    LOSS_NAME = "loss"
    Y_NAME = "y"

    def __init__(
        self,
        sequence: typing.List[nn.Module],
        loss: ThLoss,
        optim_factory: OptimFactory,
        theta_reduction: str = "mean",
    ):
        super().__init__()
        self._sequence = nn.Sequential(*sequence)
        self._loss = loss
        self._theta_step = GradStepTheta(self, optim_factory, theta_reduction)
        self._x_step = GradStepX()

    def assess_y(self, y: IO, t: IO, reduction_override: str = None) -> AssessmentDict:
        assessment = self._loss.assess_dict(y[0], t[0], reduction_override)
        assessment[self.VALIDATION_NAME] = assessment[self.LOSS_NAME]
        return assessment

    def step(self, conn: Conn, state: State, from_: IO = None) -> Conn:
        return self._theta_step.step(conn, state, from_)

    def step_x(self, conn: Conn, state: State) -> Conn:
        return self._x_step.step_x(conn, state)

    def forward(self, x: IO, state: State, detach: bool = True) -> IO:
        x.freshen(False)
        y = state[self, self.Y_NAME] = IO(self._sequence(*x.vals), detach=False)
        return y.out(detach)


class GradLoopLearner(LearningMachine, BatchIdxStepX, BatchIdxStepTheta):
    """Gradient learner designed for multiple loops"""

    LOSS_NAME = "loss"
    VALIDATION_NAME = "validation"
    Y_NAME = "y"

    def __init__(
        self,
        sequence: typing.List[nn.Module],
        loss: ThLoss,
        theta_optim_factory: OptimFactory,
        x_optim_factory: OptimFactory,
        theta_reduction: str = "mean",
        x_reduction: str = "mean",
    ):
        super().__init__()
        self._sequence = nn.Sequential(*sequence)
        self._loss = loss
        self._theta_step = GradLoopStepTheta(self, theta_optim_factory, theta_reduction)
        self._x_step = GradLoopStepX(self, x_optim_factory, x_reduction)

    def assess_y(self, y: IO, t: IO, reduction_override: str = None) -> AssessmentDict:
        assessment = self._loss.assess_dict(y[0], t[0], reduction_override)
        assessment[self.VALIDATION_NAME] = assessment[self.LOSS_NAME]
        return assessment

    def step(
        self, conn: Conn, state: State, from_: IO = None, batch_idx: Idx = None
    ) -> Conn:
        return self._theta_step.step(conn, state, from_, batch_idx)

    def step_x(self, conn: Conn, state: State, batch_idx: Idx = None) -> Conn:
        return self._x_step.step_x(conn, state, batch_idx)

    def forward(self, x: IO, state: State, detach: bool = True) -> IO:
        x.freshen(False)
        y = state[self, self.Y_NAME] = IO(self._sequence(*x.vals), detach=False)
        return y.out(detach)


def update_x(
    x: IO, lr: float = 1.0, detach: bool = False, zero_grad: bool = False
) -> IO:
    """Updates x by subtracting the gradient from x times the learning rate

    Args:
        x (IO): the IO to update. Grad must not be 0
        lr (float, optional): multipler to multiple the gradient by. Defaults to 1.0.
        detach (bool, optional): whether to detach the output. Defaults to False.
        zero_grad (bool, optional): whether the gradient should be set to none. Defaults to True.

    Returns:
        IO: updated x
    """
    updated = []
    for x_i in x:
        if isinstance(x_i, torch.Tensor):
            x_i = x_i - lr * x_i.grad
            if zero_grad:
                x_i.grad = None
        updated.append(x_i)
    return IO(*updated, detach=detach)



# 1st part
from abc import ABC, abstractmethod

# 3rd party
import torch

# local
from ..kaku import Assessment
from ..utils import to_signed_neg
from .core import (
    Individual,
    Population,
    binary_prob,
    select_best_feature,
    select_best_individual,
)
from .exploration import EqualsAssessmentDist


def keep_original(
    candidate: torch.Tensor,
    original: torch.Tensor,
    keep_prob: float,
    batch_equal: bool = False,
) -> torch.Tensor:
    """Will keep the original given the probability passed in with keep_prob

    Args:
        candidate (torch.Tensor): the candidate (i.e. updated) value
        original (torch.Tensor): the original value
        keep_prob (float): the probability with which to keep the features in the batch
        batch_equal (bool): whether to have updates be the same for all samples in the batch
    Returns:
        torch.Tensor: the updated tensor
    """
    shape = candidate.shape if not batch_equal else [1, *candidate.shape[1:]]

    to_keep = torch.rand(shape, device=candidate.device) > keep_prob
    return (~to_keep).float() * candidate + to_keep.float() * original


class Selector(ABC):
    @abstractmethod
    def __call__(self, population: Population) -> Individual:
        pass

    @abstractmethod
    def spawn(self) -> "Selector":
        pass


class StandardSelector(Selector):
    @abstractmethod
    def select(
        self, key: str, pop_val: torch.Tensor, assessment: Assessment
    ) -> torch.Tensor:
        pass

    def __call__(self, population: Population) -> Individual:
        if population.assessments is None:
            raise ValueError("Population has not been assessed")
        result = {}
        assessments = population.stack_assessments()
        for k, v in population:
            result[k] = self.select(k, v, assessments)
        return Individual(**result)


class SelectorDecorator(Selector):
    def __init__(self, base_selector: Selector):
        self.base_selector = base_selector

    @abstractmethod
    def decorate(
        self, key: str, individual: torch.Tensor, assessment: Assessment
    ) -> torch.Tensor:
        pass

    def __call__(self, population: Population) -> Individual:

        selected = self.base_selector(population)

        result = {}
        for k, v in selected:
            result[k] = self.decorate(k, v, selected.assessment)
        return Individual(**result)

    @abstractmethod
    def spawn(self) -> "SelectorDecorator":
        pass


class BestSelectorIndividual(StandardSelector):
    """Select the best individual in the population"""

    def select(
        self, key: str, pop_val: torch.Tensor, assessment: Assessment
    ) -> torch.Tensor:
        """Select the best individual in the population

        Args:
            key (str): the key for the value
            pop_val (torch.Tensor): the value with a population dimension
            assessment (Assessment): the assessment for the value

        Returns:
            torch.Tensor: the best individual in the population
        """
        return select_best_individual(pop_val, assessment)

    def spawn(self) -> "BestSelectorIndividual":
        return BestSelectorIndividual()


class BestSelectorFeature(StandardSelector):
    def select(
        self, key: str, pop_val: torch.Tensor, assessment: Assessment
    ) -> torch.Tensor:
        """Select the best features in the population

        Args:
            key (str): the key for the value
            pop_val (torch.Tensor): the value with a population dimension
            assessment (Assessment): the assessment for the value

        Returns:
            torch.Tensor: the best set of features in the population (uses the second dimension)
        """
        return select_best_feature(pop_val, assessment)

    def spawn(self) -> "BestSelectorFeature":
        return BestSelectorFeature()


class MomentumSelector(SelectorDecorator):
    def __init__(self, base_selector: Selector, momentum: float = None):
        """initializer

        Args:
            momentum (float, optional): Weight for previous time step Defaults to None.
            maximize (bool, optional): Whether to maximize the evaluation. Defaults to True.
        """
        super().__init__(base_selector)
        self._momentum = momentum
        self._params_updated = None
        self._keep_s = True
        self.diff = None
        self.cur = None
        self.dx = None

    def decorate(
        self, key: str, individual: torch.Tensor, assessment: Assessment
    ) -> torch.Tensor:

        if self.diff is None and self.cur is None:
            self.cur = individual
        elif self.diff is None:
            self.cur = individual
            self.diff = individual - self.cur
        else:
            self.cur = self.diff + individual
            self.diff = (individual - self.cur) + self._momentum * self.diff

        return self.cur

    def spawn(self) -> "MomentumSelector":
        return MomentumSelector(self.base_selector.spawn(), self._momentum)


class SlopeSelector(StandardSelector):
    """
    'Selects' the slope from current evaluation.
    Can be used to add to the value before 'spawning'
    """

    def __init__(self, momentum: float = None):
        if momentum is not None and momentum <= 0.0:
            raise ValueError(
                f"Momentum must be greater or equal to 0 or None, not {momentum}"
            )
        self._momentum = momentum
        self._slope = None

    def select(
        self, key: str, pop_val: torch.Tensor, assessment: Assessment
    ) -> torch.Tensor:
        # TODO: Add in momentum for slope (?)

        evaluation = assessment.value[:, :, None]
        ssx = (pop_val**2).sum(0) - (1 / len(pop_val)) * (pop_val.sum(0)) ** 2
        ssy = (pop_val * evaluation).sum(0) - (1 / len(pop_val)) * (
            (pop_val.sum(0) * evaluation.sum(0))
        )
        slope = ssy / ssx
        self._slope = (
            self._slope * self._momentum + slope
            if self._slope is not None and self._momentum is not None
            else slope
        )
        return self._slope

    def spawn(self) -> "SlopeSelector":
        return SlopeSelector(self._momentum)


class BinaryProbSelector(Selector):

    neg_count = "neg_count"
    pos_count = "pos_count"

    def __init__(self, x: str = "x") -> None:
        super().__init__()
        self.x = x

    def __call__(self, population: Population) -> Individual:

        x = population[self.x]
        assessment = population.stack_assessments()
        loss = assessment.value
        base_shape = [1] * len(x.shape)
        for i, sz in enumerate(assessment.shape):
            base_shape[i] = sz
        updated, pos_count, neg_count = binary_prob(x, loss, True)
        return Individual(x=updated, pos_count=pos_count, neg_count=neg_count)

    def spawn(self) -> "BinaryProbSelector":
        return BinaryProbSelector(self.x)


class BinaryGaussianSelector(Selector):
    """Value based selector for binary inputs. Uses the Gaussian distribution to
    either select based on a sample or the mean
    """

    def __init__(
        self,
        x: str = "x",
        zero_neg: bool = False,
        to_sample: bool = True,
        batch_equal: bool = False,
    ):
        """_summary_

        Args:
            x (str, optional): _description_. Defaults to 'x'.
            zero_neg (bool, optional): _description_. Defaults to False.
            to_sample (bool, optional): _description_. Defaults to True.
            batch_equal (bool, optional): whether . Defaults to False.
        """
        super().__init__()
        self.x = x
        self.zero_neg = zero_neg
        self.to_sample = to_sample
        self.batch_equal = batch_equal
        self.pos_assessment_calc = EqualsAssessmentDist(1)
        self.neg_assessment_calc = EqualsAssessmentDist(-1 if not zero_neg else 0)

    def __call__(self, population: Population) -> Individual:

        assessments = population.stack_assessments()
        x = population[self.x]
        if self.to_sample:
            pos = self.pos_assessment_calc.sample(assessments, x)
            neg = self.neg_assessment_calc.sample(assessments, x)
        else:
            pos = self.pos_assessment_calc.mean(assessments, x)
            neg = self.neg_assessment_calc.mean(assessments, x)
        result = (pos > neg).type_as(x)
        if not self.zero_neg:
            result = to_signed_neg(result)
        return Individual(**{self.x: result})

    def spawn(self) -> "BinaryGaussianSelector":
        return BinaryGaussianSelector(
            self.x, self.zero_neg, self.to_sample, self.batch_equal
        )



from abc import ABC, abstractmethod

# 3rd party
import torch

# local
from .machine import Idx


class FeatureLimitGen(ABC):
    @abstractmethod
    def sample_limit(self) -> Idx:
        pass

    def __call__(self) -> Idx:
        return self.sample_limit()


class RandomFeatureIdxGen(FeatureLimitGen):
    def __init__(self, n_features: int, choose_count: int):

        assert choose_count <= n_features
        self._n_features = n_features
        self._choose_count = choose_count

    @property
    def n_features(self) -> int:
        return self._n_features

    @n_features.setter
    def n_features(self, n_features: int):
        if n_features < self._choose_count:
            raise ValueError("")
        self._n_features = n_features

    @property
    def choose_count(self) -> int:
        return self._choose_count

    @choose_count.setter
    def choose_count(self, choose_count: int):
        if choose_count < self._choose_count:
            raise ValueError()
        self._choose_count = choose_count

    def sample_limit(self) -> Idx:
        return Idx(torch.randperm(self.n_features)[: self.choose_count], 1)



# flake8: noqa

from .assess import (
    LOSS_MAP,
    Assessment,
    AssessmentDict,
    Loss,
    ModLoss,
    Reduction,
    ThLoss,
    ThModLoss,
)
from .component import (
    Assessor,
    Autoencoder,
    Classifier,
    Decoder,
    Encoder,
    Learner,
    NNComponent,
    Regressor,
    SelfLearner,
)
from .layer_assess import AssessContext, DiffLayerAssessor, LayerAssessor
from .limit import FeatureLimitGen, RandomFeatureIdxGen
from .machine import (
    IO,
    BatchIdxStepTheta,
    BatchIdxStepX,
    Conn,
    EmissionStack,
    FeatureIdxStepTheta,
    FeatureIdxStepX,
    Idx,
    LayerIO,
    LearningMachine,
    NullLearner,
    StepHook,
    StepTheta,
    StepX,
    StepXHook,
    T,
    idx_conn,
    idx_io,
    idx_layer_io,
    idx_th,
    update_io,
    update_step_x,
    update_tensor,
)
from .optimize import (
    OPTIM_MAP,
    MetaOptim,
    NullOptim,
    OptimFactory,
    OptimFactoryX,
    itadaki,
)
from .state import IDable, MyState, State, StateKeyError
from .steps import IterHiddenStep, IterOutStep, Step, StepLoop, TwoLayerStep



# 1st party
import typing
from abc import abstractmethod
from functools import partial

# 3rd party
import torch.nn as nn

from .core import Func

# local
from .nodes import Info, Process


class ProcessSpawner(nn.Module):
    """Base class for spawning a process. Aims to improve usability"""

    @abstractmethod
    def from_(self, process: Process) -> Process:
        """Create a process that is linked to from another Process

        Args:
            process (Process): The process linked to

        Returns:
            Process: The resulting process
        """
        pass


class MSpawner(ProcessSpawner):
    """Spawner for an nn.Module"""

    def __init__(
        self, module: nn.Module, name: str = None, info: Info = None
    ) -> Process:
        """initializer

        Args:
            module (nn.Module): Module to spawn
            name (str, optional): Name of the module. Defaults to None.
            info (Info, optional): Information about the module. Defaults to None.
        """
        super().__init__()
        self.module = module
        self.name = name
        self.info = info or Info()

    def from_(self, process: Process) -> Process:
        """Create a process that is linked to from another Process

        Args:
            process (Process): The process linked to

        Returns:
            Process: The resulting process
        """
        return process.to(self.module, self.name, self.info)


class FSpawner(ProcessSpawner):
    """Process spawner for a function"""

    def __init__(
        self, f: typing.Callable, name: str = None, info: Info = None
    ) -> Process:
        """initializer

        Args:
            f (typing.Callable): The function to spawn
            name (str, optional): The name of the process. Defaults to None.
            info (Info, optional): Information for the process. Defaults to None.

        Returns:
            Process: The resulting process
        """
        super().__init__()
        self.f = f
        self.name = name
        self.info = info

    def from_(self, process: Process, *args, **kwargs) -> Process:
        """Create a process that is linked to from another Process

        Args:
            process (Process): The process linked to

        Returns:
            Process: The resulting process
        """
        return process.to(Func(self.f, *args, **kwargs), name=self.name, info=self.info)

    @classmethod
    def partial(
        cls, f: typing.Callable, *args, name: str = None, info: Info = None, **kwargs
    ) -> "FSpawner":
        """Create an FSpawner using partial to set the function args. Set them by passing
        in args and kwargs

        Args:
            f (typing.Callable): The function to spawn
            name (str, optional): The name of the Process. Defaults to None.
            info (Info, optional): Information for the Process. Defaults to None.

        Returns:
            FSpawner: The FSpawner instance that is created
        """
        return FSpawner(partial(f, *args, **kwargs), name, info)



# 1st party
from abc import ABC, abstractmethod

# 3rd party
import torch
import torch.nn as nn
import torch.nn.functional as nn_func


class Reversible(nn.Module):
    """Base class for reversible modules"""

    @abstractmethod
    def reverse(self, y: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError

    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError


class Null(Reversible):
    """
    Module that does not act on the inputs
    """

    def __init__(self, multi: bool = False):
        """initializer

        Args:
            multi (bool, optional): Whether the module can be reversed. Defaults to False.
        """
        super().__init__()
        if multi:
            self.forward = self.multi_forward
            self.reverse = self.multi_reverse
        else:
            self.forward = self.single_forward
            self.reverse = self.single_reverse

        self.multi = multi

    def multi_forward(self, *x: torch.Tensor):
        return x

    def single_forward(self, x: torch.Tensor):
        return x

    def multi_reverse(self, *y) -> torch.Tensor:
        return y

    def single_reverse(self, y) -> torch.Tensor:
        return y


class TargetReverser(ABC):
    """reverse the target"""

    @abstractmethod
    def reverse(self, x: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
        pass


class SequenceReversible(Reversible):
    """Reverse a sequence"""

    def __init__(self, *reversibles: Reversible):
        """initialzier

        Args:
            reversible (Reversible): The erversible layers in the sequence
        """
        super().__init__()
        self._reversibles = nn.ModuleList(*reversibles)

    def reverse(self, y: torch.Tensor) -> torch.Tensor:
        """
        Args:
            y (torch.Tensor): the output

        Returns:
            torch.Tensor: the output passed through the reversed sequence
        """
        for module in reversed(self._reversibles):
            y = module.reverse(y)
        return y

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x (torch.Tensor): The input

        Returns:
            torch.Tensor: The input passed through the sequence
        """
        for module in self._reversibles:
            x = module(x)
        return x


class SigmoidInvertable(Reversible):
    def reverse(self, y: torch.Tensor):
        return torch.log(y / (1 - y))

    def forward(self, x: torch.Tensor):
        return torch.sigmoid(x)


class SoftMaxReversible(Reversible):
    """Reversed softmax. Uses the inversion of the sigmoid to reverse
    This is not necessarily correct since the SoftMax is not invertable
    """

    def __init__(self, dim=-1):
        """initializer

        Args:
            dim (int, optional): _description_. Defaults to -1.
        """

        super().__init__()
        self._dim = dim

    def reverse(self, y: torch.Tensor) -> torch.Tensor:
        """Reverse the softmax operation by doing a sigmoid operation

        Args:
            y (torch.Tensor): the output of the layer

        Returns:
            Tensor: the inverted output
        """
        return torch.log(y / (1 - y))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Use SoftMax on the the layer

        Args:
            x (torch.Tensor): _description_

        Returns:
            Tensor: _description_
        """
        return nn_func.softmax(x, dim=self._dim)


class BatchNorm1DReversible(Reversible):
    """Invert the batch norm operation. Presently it doe not support the learnable parameters"""

    def __init__(self, n_features: int, momentum: float = 0.1):
        """initializer

        Args:
            n_features (int): The number of features to normalize
            momentum (float, optional): The momentum of the batchnorm. Defaults to 0.1.
        """
        super().__init__()
        self._momentum = momentum
        self._batch_norm = nn.BatchNorm1d(n_features, momentum=momentum)

    def reverse(self, y: torch.Tensor) -> torch.Tensor:
        """invert the BatchNorm operation

        Args:
            y (torch.Tensor): the input to the layer

        Returns:
            torch.Tensor: the inverted patch norm
        """
        return (
            y * torch.sqrt(self._batch_norm.running_var) + self._batch_norm.running_mean
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """batch normalize the input

        Args:
            x (torch.Tensor): _description_

        Returns:
            Tensor: The output of the batch norm operation
        """
        return self._batch_norm.forward(x)


class LeakyReLUInvertable(Reversible):
    """LeakyReLU that can be inverted"""

    def __init__(self, negative_slope: float = 0.01):
        super().__init__()
        assert negative_slope > 0.0
        self._negative_slope = negative_slope

    def reverse(self, y: torch.Tensor) -> torch.Tensor:
        """Reverse the LeakyReLU function

        Args:
            y (torch.Tensor): the output to the layer

        Returns:
            torch.Tensor: the inverted output
        """
        return nn_func.leaky_relu(y, 1 / self._negative_slope)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Send the LeakyReLU forward

        Args:
            x (torch.Tensor): the input to the layer

        Returns:
            torch.Tensor: the output of the layer
        """
        return nn_func.leaky_relu(x, self._negative_slope)


class ZeroToNeg1(Reversible):
    """Converts binary valued inputs so that -1 is negative, and 1 is positive"""

    def reverse(self, y: torch.Tensor):
        """Convert Negative one to zero

        Args:
            y (torch.Tensor): The tensor with negative one for negatives

        Returns:
            torch.Tensor: The tensor with zeros for negatives
        """
        return (y + 1) / 2

    def forward(self, x: torch.Tensor):
        """Convert Zero to Negative one

        Args:
            x (torch.Tensor): The tensor with zeros for negatives

        Returns:
            torch.Tensor: the tensor with Negative one for
        """
        return (x * 2) - 1


class Neg1ToZero(Reversible):
    """Converts binary valued inputs so that 0 is negative, and 1 is positive"""

    def __init__(self):
        super().__init__()
        self._neg = ZeroToNeg1()

    def reverse(self, y: torch.Tensor) -> torch.Tensor:
        """Convert Zero to Negative one

        Args:
            y (torch.Tensor): The tensor with zeros for negatives

        Returns:
            torch.Tensor: the tensor with Negative one for
        """
        return self._neg(y)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Convert Negative one to zero

        Args:
            x (torch.Tensor): The tensor with negative one for negatives

        Returns:
            torch.Tensor: The tensor with zeros for negatives
        """
        return self._neg.reverse(x)



import typing
from abc import abstractmethod
from functools import partial

import torch
from torch.utils.data import ConcatDataset, DataLoader, Dataset, Subset, TensorDataset

from ..kaku import IO
from .base import Material


class MaterialDecorator(Material):
    """Material used to decorate the output of a material"""

    def __init__(self, base_material: Material):
        """initializer

        Args:
            base_material (Material): The material to decorate
        """

        self._base_material = base_material

    @abstractmethod
    def decorate(self, item):
        pass

    def __iter__(self) -> typing.Iterator:
        """iterator

        Returns:
            typing.Iterator

        Yields:
            typing.Any: The values returned frmo base material
        """

        for x in self._base_material:
            yield self.decorate(x)

    def __len__(self) -> int:
        """
        Returns:
            int: the number of itms in the material
        """
        return len(self._base_material)


class IODecorator(MaterialDecorator):
    """Convert the value returned by the material to IO"""

    def decorate(self, item) -> typing.Tuple[IO, IO]:
        """
        Args:
            item (typing.Any): the value to decorate

        Returns:
            typing.Tuple[IO, IO]: The values added to an IO object
        """

        return tuple(IO(i) for i in item)


class DLMaterial(Material):
    """Material that wraps a DataLoader"""

    def __init__(self, dataloader_factory: typing.Callable[[], DataLoader]):
        """initializer

        Args:
            dataloader_factory (typing.Callable[[], DataLoader]): The DataLoader factory to wrap
        """
        self.dataloader_factory = dataloader_factory

    def __iter__(self) -> typing.Iterator:
        """Iterator

        Returns:
            typing.Iterator

        Yields:
            typing.Any: The output of the dataloader
        """
        for x in self.dataloader_factory():
            yield x

    @classmethod
    def load(
        self,
        dataset: typing.Union[Dataset, typing.List[Dataset]],
        batch_size: int,
        shuffle: bool = True,
        **kwargs,
    ) -> "DLMaterial":
        """Craete a DLMaterial from a dataset

        Args:
            dataset (typing.Union[Dataset, typing.List[Dataset]]): the dataset 
             or datasets to create the material for. if multiple passed they 
             swill be concatenated
            batch_size (int): the number of elements in the batch
            shuffle (bool, optional): whether to shuffle the DataLoader. 
             Defaults to True.

        Returns:
            DLMaterial: The resulting
        """
        if isinstance(dataset, typing.Iterable):
            dataset = ConcatDataset(dataset)
        return DLMaterial(partial(DataLoader, dataset, batch_size, shuffle, **kwargs))

    @classmethod
    def load_tensor(
        self,
        tensors: typing.Tuple[torch.Tensor],
        batch_size: int,
        shuffle: bool = True,
        **kwargs,
    ) -> "DLMaterial":
        """Load the DLMaterial from a set of tensors

        Args:
            tensors (typing.Tuple[torch.Tensor]): the tensors to load the material for
            batch_size (int): the batch size for the tensors
            shuffle (bool, optional): whether to shuffle the DataLoader. Defaults to True.

        Returns:
            DLMaterial: _description_
        """
        dataset = TensorDataset(*tensors)
        return DLMaterial(partial(DataLoader, dataset, batch_size, shuffle, **kwargs))

    def __len__(self) -> int:
        return len(self.dataloader_factory())


def split_dataset(
    dataset: Dataset,
    split_at: typing.Union[float, typing.List[float]],
    randomize: bool = True,
    seed: int = None,
) -> typing.List[Subset]:
    """split dataset into multiple datasets

    Args:
        dataset (Dataset): _description_
        split_at (typing.Union[float, typing.List[float]]): _description_
        randomize (bool, optional): _description_. Defaults to True.
        seed (int, optional): _description_. Defaults to None.

    Raises:
        ValueError: If the value for the split is invalid

    Returns:
        typing.List[Subset]: The datasets split into subsets
    """

    g = torch.Generator()
    if seed is not None:
        g.manual_seed(seed)

    if not isinstance(split_at, typing.Iterable):
        split_at = [split_at]
    if randomize:
        indices = torch.randperm(len(dataset))
    else:
        indices = torch.arange(0, len(dataset))

    prev_split = 0.0
    prev_index = 0
    size = len(dataset)
    splits = []
    for cur_split in split_at:

        if isinstance(cur_split, float):
            cur_index = int(cur_split * size)
        else:
            cur_index = cur_split

        if cur_index <= prev_index or cur_index >= size:
            raise ValueError(
                f"Invalid value for split. It must be in range ({prev_split}, 1.0]"
            )

        splits.append(Subset(dataset, indices[prev_index:cur_index]))
        prev_index = cur_index
    splits.append(Subset(dataset, indices[prev_index:]))
    return splits



# 1st party
import typing
from uuid import uuid4

import numpy as np

# 3rd party
import pandas as pd

# local
from ..kaku import AssessmentDict
from .base import Material


class Entry(object):
    def __init__(
        self,
        n_items: int,
        entry_index: int,
        info: typing.Dict = None,
        index_name: str = "epoch iteration",
        entry_index_name: str = "epoch",
    ):
        """An entry in a log. Typically use to store epoch results

        Args:
            n_items (int): The number of items that will be added to the entry
            entry_index (int): The index for the entry
            info (typing.Dict, optional): Other info about the entry. Defaults to None.
            index_name (str, optional): The name for the index column. Defaults to "epoch iteration".
            entry_index_name (str, optional): The name for the entry index column. Defaults to "epoch".
        """
        super().__init__()
        self._id = str(uuid4())
        self.info = info or {}
        self._data = []
        self.n_items = n_items
        self.index_name = index_name
        self.entry_index = entry_index
        self.entry_index_name = entry_index_name

    def __getitem__(self, idx: int):
        """
        Args:
            idx (int): Index to retrieve data for

        Returns:
            typing.Any: the value at that index
        """
        return self._data[idx]

    def __iter__(self) -> typing.Iterator:
        """
        Returns:
            typing.Iterator: Iterator to loop over

        Yields:
            typing.Any: The value at the index
        """
        for datum in self._data:
            yield datum

    def __len__(self) -> int:
        """
        Returns:
            int: the number of items in the entry
        """
        return len(self._data)

    def add(self, datum: typing.Dict):
        """Add an entry

        Args:
            datum (typing.Dict): _description_
        """
        self._data.append(datum)

    @property
    def id(self) -> str:
        return self._id

    @property
    def data(self) -> typing.List:
        """
        Returns:
            typing.List: The data stored in the entry
        """
        return list(self._data)

    @property
    def df(self) -> pd.DataFrame:
        """
        Returns:
            pd.DataFrame: The data converted to a dataframe
        """
        df = pd.DataFrame(self._data)
        df["log_id"] = self._id
        df[self.index_name] = pd.Series(range(len(df)))
        df[list(self.info.keys())] = list(self.info.values())
        df[self.entry_index_name] = self.entry_index_name
        return df


class Log(object):
    """Container for entries for a given teacher"""

    def __init__(
        self,
        teacher: str,
        info: typing.Dict = None,
        iteration_index_name: str = "iteration",
    ):
        """initializer

        Args:
            teacher (str): The name of the teacher for the log
            info (typing.Dict, optional): Info for the log. Defaults to None.
            iteration_index_name (str, optional): The name for the column for the iteration. Defaults to "iteration".
        """
        self._teacher = teacher
        self._entries: typing.List[Entry] = []
        self.info = info or {}
        self.iteration_index_name = iteration_index_name

    def add_entry(self, n_items: int, info: typing.Dict = None) -> int:
        """Add an entry to the Log

        Args:
            n_items (int): The number of items that will be added to the entry
            info (typing.Dict, optional): The info describing the entry. Defaults to None.

        Returns:
            int: The index for the entry
        """

        self._entries.append(Entry(n_items, len(self._entries), info))
        return len(self._entries) - 1

    def update_entry(self, id: int, data: typing.Dict):
        """Add data to an entry

        Args:
            id (int): The index (id) for the entry
            data (typing.Dict): The data to add to the entry
        """
        self._entries[id].add(data)

    @property
    def teacher(self) -> str:
        """
        Returns:
            str: The name of the teacher writing the log
        """
        return self._teacher

    @property
    def entries(self) -> typing.List[Entry]:
        """
        Returns:
            typing.List[Entry]: The entries in the log
        """
        return list(self._entries)

    @property
    def current(self) -> Entry:
        """
        Returns:
            Entry: The current (last) entry
        """
        return self._entries[-1]

    @property
    def df(self) -> pd.DataFrame:
        """
        Returns:
            pd.DataFrame: The data in the log converted to a DataFrame
        """
        df = pd.concat([entry.df for entry in self._entries], ignore_index=True)
        df["teacher"] = self._teacher
        df[self.iteration_index_name] = pd.Series((range(len(df))))

        df[list(self.info.keys())] = list(self.info.values())
        return df

    def __len__(self) -> int:
        """
        Returns:
            int: The number of entries in the log
        """
        return len(self._entries)


class Record(object):
    """Container for logs"""

    def __init__(self):
        """initializer"""
        self._logs: typing.Dict[str, Log] = {}

    def create_logger(self, teacher_name: str, material: Material) -> "Logger":
        """Create a logger for a teacher (Review and see if this is necessary)

        Args:
            teacher_name (str): Name of the teacher
            material (Material): The material to log for

        Returns:
            Logger
        """
        return Logger(self, teacher_name, material)

    def add_entry(
        self, teacher: str, n_items: int, info: typing.Dict[str, typing.Any] = None
    ) -> int:
        """add an entry to the record

        Args:
            teacher (str): The name of the teacher for the entry
            n_items (int): The number of items that will be added to the entry
            info (typing.Dict[str, typing.Any], optional): Supplementary info for the entry. Defaults to None.

        Returns:
            int: the index (id) for the entry
        """

        if teacher not in self._logs:
            self._logs[teacher] = Log(teacher)
        log = self._logs[teacher]
        id = log.add_entry(n_items, info)
        return id

    def update_entry(
        self,
        teacher_name: str,
        entry_id: str,
        data: typing.Union[typing.Dict[str, typing.Any], AssessmentDict],
    ):
        """add data to an entry

        Args:
            teacher_name (str): the name of the teacher to update
            entry_id (str): The index id for the entry
            data (typing.Union[typing.Dict[str, typing.Any], AssessmentDict]): The data to add to the entry
        """
        # do not want to store as a torch tensor. Not as good for DataFrames and may have
        # a graph which eats up a lot of memory
        if isinstance(data, AssessmentDict):
            data = data.numpy()

        self._logs[teacher_name].update_entry(entry_id, data)

    def current(self, teacher_name: str) -> Entry:
        """
        Args:
            teacher_name (str): The teacher to get the current entry for

        Returns:
            Entry: The current entry for the teacher
        """

        return self._logs[teacher_name].current

    def df(self, teacher_names: typing.Iterable[str] = None) -> pd.DataFrame:
        """Convert the record to a DataFrame
        Args:
            teacher_names (typing.Iterable[str], optional): The names of the teacher's to retrieve for.
                If None will retrieve for all teachers. Defaults to None.

        Returns:
            pd.DataFrame: The concatenated DataFrames
        """

        if teacher_names is not None:
            teacher_names = set(teacher_names)

        return pd.concat(
            [
                log.df
                for log in self._logs.values()
                if teacher_names is None or (log.teacher in teacher_names)
            ],
            ignore_index=True,
        )

    def __getitem__(self, teacher_name: str) -> Log:
        """
        Args:
            teacher_name (str): The teacher to retrieve a log for

        Returns:
            Log: The Log for the teacher
        """
        return self._logs[teacher_name]


class Logger(object):
    def __init__(self, record: Record, teacher: str, material: Material):
        """Convenience class to make updating a log easier

        Args:
            record (Record): The record to update
            teacher (str): The name of the teacher to update for
            material (Material): The material used to teach with
        """
        self._teacher = teacher
        self._record = record
        self._material = material
        self._entry_id = record.add_entry(teacher, len(material))

    def __call__(
        self, data: typing.Union[typing.Dict[str, typing.Any], AssessmentDict]
    ) -> typing.Union[typing.Dict, AssessmentDict]:
        """Update the current entry for the teacher

        Args:
            data (typing.Union[typing.Dict[str, typing.Any], AssessmentDict]): The data

        Returns:
            typing.Union[typing.Dict, AssessmentDict]: The data passed into the argument data
        """

        self._record.update_entry(
            self._teacher,
            self._entry_id,
            data.numpy() if isinstance(data, AssessmentDict) else data,
        )
        return data

    def advance(self, new_material: Material = None):
        """Advance the current Log by adding a new entry
        Args:
            new_material (Material, optional): The material for the new entry. Defaults to None if not updated.
        """
        if new_material is not None:
            self._material = new_material
        self._log_id = self._record.add_entry(self._teacher, len(self._material))

    @property
    def record(self) -> Record:
        """
        Returns:
            Record: The record for the Logger
        """
        return self._record

    @property
    def log(self) -> Log:
        """
        Returns:
            Log: The log for the teacher
        """
        return self._record[self._teacher]

    @property
    def df(self) -> pd.DataFrame:
        """
        Returns:
            pd.DataFrame: The DataFrame for the Teacher
        """
        return self._record[self._teacher].df


class Results(object):
    """Container for training or testing results. Mainly used for showing progress"""

    def __init__(self, window: int = None):
        """initializer

        Args:
            window (int, optional): The number of results to get the mean for. Defaults to None.
        """

        self.window = window
        self._results: typing.Dict[str, typing.List] = {}

    def add(self, result: typing.Dict[str, np.ndarray]):
        """Add result to the container

        Args:
            result (typing.Dict[str, np.ndarray]): The result to add
        """

        for k, v in result.items():
            if k not in self._results:
                self._results[k] = []
            self._results[k].append(v.item())

    def __getitem__(self, key: str) -> typing.List:
        """
        Args:
            key (str): The key to retrieve for

        Returns:
            typing.List: the result specified by key (e.g. "loss")
        """
        return list(self._results[key])

    def aggregate(
        self,
        result_choice: typing.Union[str, typing.List[str]] = None,
        window_override: int = None,
    ) -> typing.Dict[str, float]:
        """Aggregate the results

        Args:
            result_choice (typing.Union[str, typing.List[str]], optional): Keys 
              to retrieve results for. Defaults to None.

        Returns:
            typing.Dict[str, float]: a dictionary of the results
        """
        window = window_override or self.window
        result_choice = result_choice or list(self._results.keys())
        if isinstance(result_choice, str):
            result_choice = [result_choice]

        aggregated = {}
        for k in result_choice:
            result = self._results[k]
            if window is not None:
                result = result[-window:]
            aggregated[k] = np.mean(result)
        return aggregated



"""
Modules to implement exploration
on the forward pass
"""

import typing
from abc import ABC, abstractmethod

# 1st party
from typing import Any

# 3rd party
import torch
import torch.nn as nn

# local
from ..kaku import IO, Assessment
from ..utils import get_model_parameters, update_model_parameters
from .core import gather_idx_from_population, gaussian_sample

# TODO: Consider how to handle these
# Probably get rid of the first


class NoiseReplace(torch.autograd.Function):
    """
    Replace x with a noisy value. The gradInput for x will be the gradOutput and
    for the noisy value it will be x

    Note: May cause problems if only evaluating on a subset of outputs.
    The gradient may be 0 but in that case so it will set the target to
    be "noise" which is likely undesirable. In that case, use NoiseReplace2
    """

    @staticmethod
    def forward(ctx, x, noisy):
        ctx.save_for_backward(x, noisy)
        return noisy.clone()

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):

        x, noisy = ctx.saved_tensors
        return (noisy + grad_output) - x, None


class NoiseReplace2(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, noisy):
        ctx.save_for_backward(x, noisy)
        return noisy

    @staticmethod
    def backward(ctx, grad_output):

        x, noisy = ctx.saved_tensors
        grad_input = (noisy + grad_output) - x
        direction = torch.sign(grad_input)
        magnitude = torch.min(grad_output.abs(), grad_input.abs())
        return direction * magnitude, None


class NoiseReplace3(torch.autograd.Function):
    """
    Replace x with a noisy value. The gradInput for x will be the gradOutput and
    for the noisy value it will be x.

    Uses kind of a hack with 'chosen_idx' so that all entries that are not chosen
    will be zero
    """

    @staticmethod
    def forward(ctx, x, noisy, chosen_idx):
        ctx.save_for_backward(x, noisy)
        return noisy.clone(), chosen_idx.clone()

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor, chosen_idx):

        x, noisy = ctx.saved_tensors
        # grad_input_ = grad_input.gather(1, chosen_idx)
        # noisy = noisy.gather(1, chosen_idx)
        # x = x.gather(1, chosen_idx)
        # grad_output = grad_output.gather(1, chosen_idx)
        grad_input_base = (noisy + grad_output) - x

        grad_input_base = grad_input_base.view(
            chosen_idx.shape[0], chosen_idx.shape[1], -1
        )
        chosen_idx_cur = chosen_idx[:, :, None].repeat(1, 1, grad_input_base.shape[2])

        grad_input_zeros = torch.zeros_like(grad_input_base)
        grad_input_zeros.scatter_(
            1, chosen_idx_cur, grad_input_base.gather(1, chosen_idx_cur)
        )

        return grad_input_zeros.view(grad_output.shape), None, chosen_idx


class ChooseIdx(torch.autograd.Function):
    """
    Use with NoiseReplace3

    This chooses an index so that on the backpropagation only
    """

    @staticmethod
    def forward(ctx, x, chosen_idx):
        ctx.save_for_backward(chosen_idx)
        return x

    @staticmethod
    def backward(ctx: Any, grad_output: torch.Tensor) -> Any:
        (chosen_idx,) = ctx.saved_tensors
        return grad_output, chosen_idx


class ExplorerNoiser(nn.Module):
    """Add noise to the input"""

    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pass


class GaussianNoiser(ExplorerNoiser):
    """Add Gaussian noise to the exploration"""

    def __init__(self, std: float = 1.0, mu: float = 0.0):
        super().__init__()
        self.std = std
        self.mu = mu

    def forward(self, x: torch.Tensor):
        return (
            torch.randn(x.size(), dtype=x.dtype, device=x.device) * self.std + self.mu
        )


class ExplorerSelector(nn.Module):
    @abstractmethod
    def forward(self, x: torch.Tensor, noisy: torch.Tensor):
        pass


class RandSelector(ExplorerSelector):
    """Randomly choose whether to select the noisy value or the original x"""

    def __init__(self, select_noise_prob: float):
        super().__init__()
        self.select_noise_prob = select_noise_prob

    def forward(self, x: torch.Tensor, noisy: torch.Tensor):
        assert noisy.size() == x.size()
        selected_noise = (
            torch.rand(noisy.size(), device=x.device) <= self.select_noise_prob
        )
        return (
            selected_noise.type_as(noisy) * noisy + (~selected_noise).type_as(noisy) * x
        )


class Explorer(nn.Module):
    def __init__(self, noiser: ExplorerNoiser, selector: ExplorerSelector):
        super().__init__()
        self._noiser = noiser
        self._selector = selector

    def forward(self, x: torch.Tensor):

        with torch.no_grad():
            noisy = self._selector(x, self._noiser(x))
        return NoiseReplace.apply(x, noisy)


def remove_noise(
    x: torch.Tensor, x_noisy: torch.Tensor, k: int, remove_idx: int = 0
) -> torch.Tensor:
    """Remove noise at specified index. Assumes that the trials are in dimension 0

    Args:
        x (torch.Tensor): The original tensor
        x_noisy (torch.Tensor): The tensor with noise added to it
        k (int): The
        remove_idx (int, optional): _description_. Defaults to 0.

    Returns:
        torch.Tensor: noisy tensor with value at specified indexed replaced with non-noisy version
    """
    original_shape = x_noisy.shape
    new_shape = torch.Size([k, x.shape[0] // k, *x.shape[1:]])
    x = x.reshape(new_shape)
    x_noisy = x_noisy.reshape(new_shape)
    x_noisy[remove_idx] = x[remove_idx]
    # mask_x = torch.zeros_like(x)
    # mask_noisy = torch.ones_like(x_noisy)
    # mask_x[remove_idx] = 1
    # mask_noisy[remove_idx] = 0
    # x_noisy = mask_x * x + mask_noisy * x_noisy
    return x_noisy.reshape(original_shape)


def expand_k(x: torch.Tensor, k: int, reshape: bool = True) -> torch.Tensor:
    """expand the trial dimension in the tensor (separates the trial dimension from the sample dimension)

    Args:
        x (torch.Tensor): The tensor to update
        k (int): The number of trials
        reshape (bool, optional): Whether to use reshape (True) or view (False). Defaults to True.

    Returns:
        torch.Tensor: The expanded tensor
    """
    shape = torch.Size([k, -1, *x.shape[1:]])
    if reshape:
        return x.reshape(shape)
    return x.view(shape)


def collapse_k(x: torch.Tensor, reshape: bool = True) -> torch.Tensor:
    """collapse the trial dimension in the tensor (merges the trial dimension with the sample dimension)

    Args:
        x (torch.Tensor): The tensor to update
        reshape (bool, optional): Whether to use reshape (True) or view (False). Defaults to True.

    Returns:
        torch.Tensor: The collapsed tensor
    """
    if reshape:
        return x.reshape(-1, *x.shape[2:])
    return x.view(-1, *x.shape[2:])


class Indexer(object):
    """"""

    def __init__(self, idx: torch.LongTensor, k: int, maximize: bool = False):
        """initializer

        Args:
            idx (torch.LongTensor): index the tensor
            k (int): the number of samples in the population
            maximize (bool, optional): Whether to maximize or minimize. Defaults to False.
        """
        self.idx = idx
        self.k = k
        self.maximize = maximize
        # self.spawner = spawner

    def index(self, io: IO, detach: bool = False):
        ios = []
        for io_i in io:
            io_i = io_i.view(self.k, -1, *io_i.shape[1:])
            ios.append(gather_idx_from_population(io_i, self.idx)[0])
        return IO(*ios, detach=detach)


class RepeatSpawner(object):
    """"""

    def __init__(self, k: int):
        self.k = k

    def __call__(self, x: torch.Tensor):

        return (
            x[None]
            .repeat(self.k, *([1] * len(x.shape)))
            .reshape(self.k * x.shape[0], *x.shape[1:])
        )

    def spawn_io(self, io: IO):
        """

        Args:
            io (IO): _description_

        Returns:
            _type_: _description_
        """
        xs = []
        for x in io:
            if isinstance(x, torch.Tensor):
                x = self(x)
            xs.append(x)
        return IO(*xs)

    def select(self, assessment: Assessment) -> typing.Tuple[Assessment, Indexer]:
        """

        Args:
            assessment (Assessment):

        Returns:
            typing.Tuple[Assessment, Indexer]:
        """
        assert assessment.value.dim() == 1
        expanded = expand_k(assessment.value, self.k, False)
        if assessment.maximize:
            value, idx = expanded.max(dim=0, keepdim=True)
        else:
            value, idx = expanded.min(dim=0, keepdim=True)
        return Assessment(value, assessment.maximize), Indexer(
            idx, self.k, assessment.maximize
        )


class ModuleNoise(nn.Module):
    """Use to add noise to the model that is dependent on the direction that the model is moving in"""

    def __init__(self, module_clone: nn.Module, n_instances: int, weight: float = 0.1):
        super().__init__()
        if not (0.0 < weight < 1.0):
            raise ValueError("Weight must be in range (0, 1)")
        self._module_clone = module_clone
        self._weight = weight
        self._parameters = get_model_parameters(module_clone)
        self._direction_mean = torch.zeros_like(self._parameters)
        self._direction_std = torch.zeros_like(self._parameters)
        self._updated = False
        self._n_instances = n_instances

    def update(self, base_module):
        parameters = get_model_parameters(base_module)
        dp = parameters - self._parameters
        self._direction_var = (
            1 - self._weight
        ) * self._direction_var + self._weight * (dp - self._direction_mean) ** 2

        if self._updated:
            self._direction_mean = (
                1 - self._weight
            ) * self._direction_mean + self._weight * (dp)
        else:
            self._direction_mean = dp

        self._updated = True

    def forward(self, x: torch.Tensor):
        x = x.view(self._n_instances, -1, *x.shape)
        ps = (
            torch.randn(1, *self._direction_mean.shape, dtype=x.dtype, device=x.device)
            * torch.sqrt(self._direction_var[None])
            + self._direction_mean[None]
        )
        for x_i, p_i in (x, ps):
            update_model_parameters(self._module_clone, p_i)
            ys = self._module_clone(x_i)
        ys = torch.vstack(ys)
        return ys.view(ys.shape[0] * ys.shape[1], *ys.shape)


class AssessmentDist(ABC):
    @abstractmethod
    def __call__(
        self, assessment: Assessment, x: torch.Tensor
    ) -> typing.Union[torch.Tensor, torch.Tensor]:
        """_summary_

        Args:
            assessment (Assessment): the assessment. Must be of dimension [k, batch]
            x (torch.Tensor): the input to assess. must be of dimension
              [k, batch, feature]

        Returns:
            typing.Union[torch.Tensor, torch.Tensor]:
              The mean of the assessment, the standard deviation of the
              assessment
        """
        pass


class EqualsAssessmentDist(AssessmentDist):
    """Determine the distribution of the assessment to draw samples 
    or get the mean"""

    def __init__(self, equals_value):

        self.equals_value = equals_value

    def __call__(self, assessment: Assessment, x: torch.Tensor) -> torch.Tensor:

        value = assessment.value[:, :, None]
        if value.dim() != 3:
            raise ValueError("Value must have dimension of 3 ")
        if x.dim() != 3:
            raise ValueError("Argument x must have dimension of 3")
        equals = (x == self.equals_value).type_as(x)
        value_assessment = (equals).type_as(x) * value
        var = value_assessment.var(dim=0)
        weight = x.shape[0] / equals.sum(dim=0)
        return (
            weight * value_assessment.mean(dim=0),
            torch.sqrt(weight * var + 1e-8),
        )

    def sample(
        self, assessment: Assessment, x: torch.Tensor, n_samples: int = None
    ) -> torch.Tensor:
        mean, std = self(assessment, x)
        return gaussian_sample(mean, std, n_samples)

    def mean(self, assessment: Assessment, x: torch.Tensor) -> torch.Tensor:
        mean, _ = self(assessment, x)
        return mean



# 1st party
import typing
from abc import ABC, abstractmethod, abstractproperty

# local
from .assess import Assessment, AssessmentDict
from .machine import Conn, LearningMachine


class LayerAssessor(ABC):
    """Base class for assessing a layer"""

    CHOICES = set()

    def __init__(self):
        self._registry: typing.Dict[
            str, typing.Tuple[LearningMachine, LearningMachine]
        ] = {}
        self._assessments = None
        self._retrieve_choice = {}

    def register(
        self,
        name: str,
        machine: LearningMachine,
        outgoing: LearningMachine,
        retrieve_choice: typing.Iterable[str] = None,
    ):
        retrieve_choice = retrieve_choice or self.CHOICES
        retrieve_choice = set(retrieve_choice)
        if len(retrieve_choice.difference(self.CHOICES)) != 0:
            raise ValueError(
                f"Retrieve choice must be from {self.CHOICES} not {retrieve_choice}"
            )
        self._registry[name] = (machine, outgoing)
        self._retrieve_choice[name] = retrieve_choice

    @abstractmethod
    def update_before(self, name: str, conn: Conn):
        pass

    @abstractmethod
    def update_after(self, name: str, conn: Conn):
        pass

    def layer_assess(self, name: str, conn: Conn):
        return AssessContext(self, name, conn)

    @abstractproperty
    def assessment_dict(self) -> AssessmentDict:
        pass


class DiffLayerAssessor(LayerAssessor):
    """ """

    CHOICES = set(["incoming", "full", "outgoing"])

    def __init__(
        self,
        prefix: str,
        weight: float = 0.2,
        loss_name: str = "loss",
    ):
        super().__init__()
        self._before: typing.Dict[str, typing.Dict[Assessment]] = {}
        self._after: typing.Dict[str, typing.Dict[Assessment]] = {}
        self._diff: typing.Dict[str, typing.Dict[str, Assessment]] = {}
        self.weight = weight
        self.loss_name = loss_name
        self.prefix = prefix

    def _calc_assessments(
        self, name: str, conn: Conn
    ) -> typing.Dict[str, AssessmentDict]:

        machine, outgoing = self._registry[name]
        status = machine.training
        machine.train(False)
        if outgoing is not None:
            outgoing_status = outgoing.training
            outgoing.train(False)
        y = machine(conn.step.x)
        retrieve_choice = self._retrieve_choice[name]
        retrieved = {}

        if "incoming" in retrieve_choice:
            retrieved["incoming"] = (
                machine.assess_y(y, conn.step.t, "mean")[self.loss_name].detach().cpu()
            )
        if "full" in retrieve_choice:
            if outgoing is None:
                raise ValueError("Cannot calculate assessment as outgoing is none")
            retrieved["full"] = (
                outgoing.assess(y, conn.step_x.t, "mean")[self.loss_name].detach().cpu()
            )
        if "outgoing" in retrieve_choice:
            if outgoing is None:
                raise ValueError("Cannot calculate assessment as outgoing is none")
            retrieved["outgoing"] = (
                outgoing.assess(conn.step_x.x, conn.step_x.t, "mean")[self.loss_name]
                .detach()
                .cpu()
            )

        if outgoing is not None:
            outgoing.train(outgoing_status)
        machine.train(status)
        return retrieved

    def update_before(self, name: str, conn: Conn):
        self._before[name] = self._calc_assessments(name, conn)
        self._after[name] = {}

    def update_after(self, name: str, conn: Conn):
        if len(self._after[name]) != 0:
            raise ValueError(
                "Must call 'update_before' before calling 'update after'"
            )
        self._after[name] = self._calc_assessments(name, conn)

        diff = {}
        for k, b_i in self._before[name].items():
            a_i = self._after[name][k]
            diff[k] = Assessment((a_i.value - b_i.value))
        self._diff[name] = diff

    def layer_assess(self, name: str, conn: Conn):
        return AssessContext(self, name, conn)

    def _to_dict(self):

        results = {}

        for name in self._after.keys():

            for k, v in self._after[name].items():
                results[f"{self.prefix}_{name}_{k}_after"] = v

        for name in self._before.keys():
            for k, v in self._before[name].items():
                results[f"{self.prefix}_{name}_{k}_before"] = v
        for name in self._diff.keys():
            for k, v in self._diff[name].items():
                results[f"{self.prefix}_{name}_{k}"] = v
        return results

    @property
    def assessment_dict(self) -> AssessmentDict:
        if self._diff is None:
            return AssessmentDict()

        # results = {}
        # for k in self._diff.keys():

        #     results.update(self._to_dict(k))

        return AssessmentDict(**self._to_dict())


class AssessContext(object):
    def __init__(self, layer_assessor: LayerAssessor, name: str, conn: Conn):

        self.layer_assessor = layer_assessor
        self.name = name
        self.conn = conn

    def __enter__(self):
        self.layer_assessor.update_before(self.name, self.conn)
        return self

    def __exit__(self, type, value, traceback):
        if type is not None:
            return False
        self.layer_assessor.update_after(self.name, self.conn)
        return True



# 1st party
import math
import typing

# 3rd party
import torch
from torch import nn

# local
from . import utils
from .reversibles import Reversible


class Stride2D(nn.Module):
    def __init__(
        self,
        in_features: int,
        kernel_size: typing.Tuple[int],
        stride: typing.Tuple[int] = None,
        padding: typing.Tuple[int] = None,
        reshape_output: bool = True,
    ):

        super().__init__()
        stride = stride or 1
        padding = padding or 0
        self._in_features = in_features
        self._kernel_size = utils.to_2dtuple(kernel_size)
        self._stride = utils.to_2dtuple(stride)
        if padding is not None:
            padding = utils.to_2dtuple(padding)
            self._padding = tuple(
                map(int, (padding[0], padding[0], padding[1], padding[1]))
            )
            # self._padding = tuple(map(int, (
            #     math.floor(padding[0] / 2),
            #     math.ceil(padding[0] / 2),
            #     math.floor(padding[1] / 2),
            #     math.ceil(padding[1] / 2)
            # )))
        else:
            self._padding = None
        self._reshape_output = reshape_output

    @property
    def out_features(self) -> int:
        return self._kernel_size[0] * self._kernel_size[1] * self._in_features

    @property
    def out_shape(self) -> typing.Tuple[int, int, int]:
        return (self._in_features, *self._kernel_size)

    def calc_size(self, in_width: int, in_height: int) -> typing.Tuple[int, int]:
        """Calc the width and height after processeing

        Args:
            in_width (int): The width of the image input
            in_height (int): The height of the image output

        Returns:
            typing.Tuple[int, int]: The width and height after processing. These will
            become a part of the batch dimension
        """
        return (
            (in_width - self._kernel_size[0] + 1 + self._padding[0] + self._padding[1])
            // self._stride[0],
            (in_height - self._kernel_size[1] + 1 + self._padding[2] + self._padding[3])
            // self._stride[1],
        )

    def forward(self, x: torch.Tensor):
        if self._padding is not None:
            x = torch.nn.functional.pad(x, self._padding, "constant", 0)

        strided = torch.as_strided(
            x,
            utils.calc_size2d(x, self._stride, self._kernel_size),
            utils.calc_stride2d(x, self._stride),
        )
        strided = strided.permute(0, 2, 3, 1, 4, 5)
        if not self._reshape_output:
            return strided
        return strided.reshape(-1, math.prod(strided.shape[3:]))


class UndoStride2D(Reversible):
    """ """

    def __init__(self, n_channels: int, size: typing.Tuple[int]):
        super().__init__()
        self._n_channels = n_channels
        self._size = utils.to_2dtuple(size)

    @property
    def n_channels(self) -> int:
        return self._n_channels

    @property
    def out_shape(self) -> typing.Tuple[int, int, int]:
        return self._n_channels, self._size[0], self._size[1]

    def forward(self, x: torch.Tensor):

        return x.reshape(-1, self._size[0], self._size[1], self._n_channels).permute(
            0, 3, 1, 2
        )

    def reverse(self, y: torch.Tensor) -> torch.Tensor:
        return y.permute(0, 2, 3, 1).reshape(-1, y.size(1))


class TargetStride(nn.Module):
    """Convert the target to match a strided output"""

    def __init__(self, out_channels: int, width: int, height: int):
        super().__init__()

        self.out_channels = out_channels
        self.width = width
        self.height = height

    def forward(self, x: torch.Tensor):

        return (
            x.view(x.size(0), self.out_channels, self.width, self.height)
            .permute(0, 2, 3, 1)
            .reshape(-1, self.out_channels)
        )



# flake8: noqa

from .assessors import PopulationAssessor, XPopulationAssessor
from .core import (
    Individual,
    Population,
    binary_prob,
    cat_params,
    deflatten,
    expand,
    expand_t,
    flatten,
    gather_idx_from_population,
    gaussian_sample,
    reduce_assessment_dim0,
    reduce_assessment_dim1,
    select_best_feature,
    select_best_individual,
)
from .exploration import (
    AssessmentDist,
    ChooseIdx,
    EqualsAssessmentDist,
    Explorer,
    ExplorerNoiser,
    ExplorerSelector,
    GaussianNoiser,
    Indexer,
    ModuleNoise,
    NoiseReplace,
    NoiseReplace2,
    NoiseReplace3,
    RandSelector,
    RepeatSpawner,
    collapse_k,
    expand_k,
    remove_noise,
)
from .modifiers import (
    BinaryAdjGaussianModifier,
    BinaryGaussianModifier,
    BinaryProbModifier,
    KeepModifier,
    PopulationModifier,
    SelectionModifier,
    SlopeModifier,
)
from .populators import (
    BinaryPopulator,
    BinaryProbPopulator,
    ConservativePopulator,
    GaussianPopulator,
    PerceptronProbPopulator,
    PopulationLimiter,
    Populator,
    PopulatorDecorator,
    RepeatPopulator,
    SimpleGaussianPopulator,
    StandardPopulator,
    populate_t,
)
from .selectors import (
    BestSelectorFeature,
    BestSelectorIndividual,
    BinaryGaussianSelector,
    BinaryProbSelector,
    MomentumSelector,
    Selector,
    SelectorDecorator,
    SlopeSelector,
    StandardSelector,
    keep_original,
)



# 1st party
import typing
from abc import ABC, abstractmethod

import numpy as np
import scipy.linalg

# 3rd party
import torch
import torch.nn as nn

from ..kaku.machine import (
    IO,
    AssessmentDict,
    Conn,
    LearningMachine,
    State,
    StepTheta,
    StepX,
    ThLoss,
    update_io,
)

# local
from ..utils import to_np, to_th_as


# TODO: Move to itadaki
class LeastSquaresSolver(ABC):
    @abstractmethod
    def solve(self, a: torch.Tensor, b: torch.Tensor):
        pass


class LeastSquaresStandardSolver(LeastSquaresSolver):
    """Solve least squares"""

    def __init__(self, bias: bool = False):
        """initializer

        Args:
            bias (bool, optional): Whether there is a bias. Defaults to False.
        """

        if bias:
            self._prepare = self._prepare_with_bias
        else:
            self._prepare = self._prepare_without_bias

    def _prepare_without_bias(
        self, a: np.ndarray, b: np.ndarray
    ) -> typing.Tuple[np.ndarray, np.ndarray]:
        return a, b

    def _prepare_with_bias(
        self, a: np.ndarray, b: np.ndarray
    ) -> typing.Tuple[np.ndarray, np.ndarray]:
        m, _ = np.shape(a)
        a = np.hstack([a, np.ones((m, 1))])
        return a, b

    def solve(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        """Solve least squares between a and b

        Args:
            a (torch.Tensor): input
            b (torch.Tensor): target

        Returns:
            torch.Tensor: the least squares solution
        """
        new_, _, _, _ = scipy.linalg.lstsq(*self._prepare(to_np(a), to_np(b)))
        return to_th_as(new_, a).T


class LeastSquaresRidgeSolver(LeastSquaresSolver):
    """Solve least squares using ridge regression"""

    def __init__(self, lam: float = 1e-1, bias: bool = False):

        self._bias = bias
        self._lambda = lam
        if self._bias:
            self._prepare = self._prepare_with_bias
        else:
            self._prepare = self._prepare_without_bias

    def _prepare_without_bias(self, a: np.ndarray, b: np.ndarray):
        _, n = np.shape(a)
        lower_half = np.zeros((n, n))
        np.fill_diagonal(lower_half, np.sqrt(self._lambda))
        return (np.vstack((a, lower_half)), np.vstack([b, np.zeros((n, b.shape[1]))]))

    def _prepare_with_bias(self, a: np.ndarray, b: np.ndarray):
        m, n = np.shape(a)
        # print('Preparing w/')
        upper_half = np.hstack([a, np.ones((m, 1))])
        lower = np.zeros((n, n))
        np.fill_diagonal(lower, np.sqrt(self._lambda))
        lower_half = np.hstack([lower, np.zeros((n, 1))])
        return (
            np.vstack((upper_half, lower_half)),
            np.vstack([b, np.zeros((n, b.shape[1]))]),
        )

    def solve(self, a: torch.Tensor, b: torch.Tensor):
        """Solve least squares between a and b

        Args:
            a (torch.Tensor): input
            b (torch.Tensor): target

        Returns:
            torch.Tensor: the least squares solution
        """
        A, B = self._prepare(to_np(a), to_np(b))
        new_, _, _, _ = scipy.linalg.lstsq(A.T @ A, A.T @ B)
        # print(new_.shape)

        return to_th_as(new_, a).T


class LeastSquaresStepTheta(StepTheta):
    def __init__(
        self,
        linear: nn.Linear,
        solver: LeastSquaresSolver = LeastSquaresStandardSolver,
        optimize_dw: bool = False,
        lr: typing.Optional[float] = None,
    ):
        """initializer

        Args:
            linear (nn.Linear): _description_
            solver (LeastSquaresSolver, optional): _description_. Defaults to LeastSquaresStandardSolver.
            optimize_dw (bool, optional): _description_. Defaults to False.
            lr (typing.Optional[float], optional): _description_. Defaults to None.
            is_fresh (bool, optional): _description_. Defaults to True.
        """
        self.linear = linear
        self.solver = solver
        self._optimize = self._optimize_dw if optimize_dw else self._optimize_w
        self._lr = lr

    def split(self, p: torch.Tensor) -> typing.Tuple[torch.Tensor, torch.Tensor]:
        if self.linear.bias is None:
            return p, None

        return p[:, :-1], p[:, -1]

    def _optimize_dw(self, x: torch.Tensor, t: torch.Tensor):
        t_delta = t - self.linear.forward(x)
        dweight, dbias = self.split(self.solver.solve(x, t_delta))
        self.linear.weight.data = self.linear.weight.data + dweight
        if dbias is not None:
            self.linear.bias.data = self.linear.bias.data + dbias

    def _optimize_w(self, x: torch.Tensor, t: torch.Tensor):
        weight, bias = self.split(self.solver.solve(x, t))
        self.linear.weight.data = weight
        if bias is not None:
            self.linear.bias.data = bias

    def step(self, conn: Conn, state: State, from_: IO = None) -> Conn:
        self._optimize(conn.step.x[0], conn.step.t[0])
        return conn.connect_in(from_)


class LeastSquaresStepX(StepX):
    def __init__(
        self,
        linear: nn.Linear,
        solver: LeastSquaresSolver = LeastSquaresStandardSolver,
        optimize_dx: bool = False,
        lr: typing.Optional[float] = None,
    ):
        """initializer

        Args:
            linear (nn.Linear): _description_
            solver (LeastSquaresSolver, optional): _description_. Defaults to LeastSquaresStandardSolver.
            optimize_dx (bool, optional): _description_. Defaults to False.
            lr (typing.Optional[float], optional): _description_. Defaults to None.
            is_fresh (bool, optional): _description_. Defaults to True.
        """
        self.linear = linear
        self.solver = solver
        self._optimize = self._optimize_dx if optimize_dx else self._optimize_x
        self._lr = lr

    def _optimize_dx(self, x: torch.Tensor, t: torch.Tensor):
        y = self.linear(x)
        if self.linear.bias is not None:
            t = t - self.linear.bias[None]
            y = y - self.linear.bias[None]
        t_delta = t - y
        dx = self.solver.solve(self.linear.weight, t_delta.T)
        return x + dx

    def _optimize_x(self, x: torch.Tensor, t: torch.Tensor):
        if self.linear.bias is not None:
            t = t - self.linear.bias[None]
        return self.solver.solve(self.linear.weight, t.T)

    def step_x(self, conn: Conn, state: State) -> Conn:
        """Update x

        Args:
            conn (Conn): The connection to update with
            state (State): The current learning state

        Returns:
            Conn: The connection
        """
        x = self._optimize(conn.step_x.x[0], conn.step_x.t[0])
        update_io(IO(x), conn.step_x.x)
        conn.tie_inp()
        return conn


class LeastSquaresLearner(LearningMachine):
    """Learner that uses least squares"""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        optimize_dx: bool = True,
        x_lr: float = 1e-2,
        x_reduction: str = "mean",
    ):
        super().__init__()
        self._linear = nn.Linear(in_features, out_features, bias)
        self._loss = ThLoss("mse", "mean")
        self._step_x = LeastSquaresStepX(
            self._linear, LeastSquaresRidgeSolver(1e-4, False), optimize_dx
        )
        self._step_theta = LeastSquaresStepTheta(
            self._linear, LeastSquaresRidgeSolver(1e-3, bias=bias), optimize_dx
        )

    def assess_y(self, y: IO, t: IO, reduction_override: str = None) -> AssessmentDict:
        return self._loss.assess_dict(y[0], t[0], reduction_override=reduction_override)

    def step(self, conn: Conn, state: State, from_: IO = None) -> Conn:
        conn = self._step_theta.step(conn, state, from_)
        return conn

    def step_x(self, conn: Conn, state: State) -> Conn:
        conn = self._step_x.step_x(conn, state)
        return conn

    def forward(self, x: IO, state: State, detach: bool = True) -> IO:
        x.freshen(False)
        return IO(self._linear(x[0]), detach=detach)



# flake8: noqa

from .ensemble import VoterEnsemble, VoterEnsembleMachine
from .estimators import (
    ScikitBinary,
    ScikitEstimator,
    ScikitLimitGen,
    ScikitMachine,
    ScikitMulticlass,
    ScikitRegressor,
)



# 1st party
import typing
from collections import deque
from copy import deepcopy

import torch.nn as nn

# 3rd party
import torch.nn.functional
from torch.nn.functional import one_hot

# local
from ..kaku import (
    IO,
    AssessmentDict,
    Conn,
    FeatureIdxStepTheta,
    FeatureIdxStepX,
    Idx,
    LearningMachine,
    Loss,
    State,
)
from . import estimators


class Voter(nn.Module):
    """Module that chooses the best"""

    def __init__(self, use_sign: bool = False, n_classes: int = None):
        """initializer

        Args:
            use_sign (bool, optional): Whether to use the sign on the output for binary results. Defaults to False.
            n_classes (int, optional): Whether the inputs are . Defaults to None.

        Raises:
            ValueError: _description_
        """

        # TODO: Add support for LongTensors by using one_hot encoding
        # I will split the voter up at that point though
        #
        super().__init__()
        self._use_sign = use_sign
        self._n_classes = n_classes
        if n_classes and use_sign:
            raise ValueError(
                "Arguments use_counts and use_sign are mutually exclusive so cannot both be true"
            )
        if self._n_classes is not None:
            raise NotImplementedError

    def forward(
        self, votes: torch.Tensor, weights: typing.List[float] = None
    ) -> torch.Tensor:

        if self._n_classes is not None:
            votes = one_hot(votes, self._n_classes).sum(dim=-2)
            # TODO: FINISH
            return votes

        if weights is not None:
            votes_ = votes.view(votes.size(0), -1)
            weights_th = torch.tensor(weights, device=votes.device)[None]
            chosen = (votes_ * weights_th).sum(dim=0) / weights_th.sum(dim=0)
            chosen = chosen.view(votes.shape[1:])
        else:
            chosen = votes.mean(dim=0)
        if self._use_sign:
            return chosen.sign()

        return chosen


class VoterEnsemble(nn.Module):
    def __init__(self, base_estimator: estimators.ScikitEstimator, n_keep: int):
        """

        Args:
            base_estimator (scikit.ScikitEstimator): _description_
            n_keep (int): The number of estimators to keep

        Raises:
            ValueError: If n_keep is less than or equal to 0
            ValueError: Estimator is incorrect
        """
        super().__init__()
        if n_keep <= 0:
            raise ValueError(f"Argument n_keep must be greater than 0 not {n_keep}.")
        self._voter = None

        self._voter = Voter(isinstance(base_estimator, estimators.ScikitBinary))
        if isinstance(self._voter, estimators.ScikitMulticlass):
            raise ValueError("Multiclass classification is not supported.")
        self._estimators = deque()
        self._base_estimator = base_estimator
        self._n_keep = n_keep
        self._sklearn_fitted = lambda: True
        self._fitted = False

    @property
    def n_keep(self) -> int:
        return self._n_keep

    @n_keep.setter
    def n_keep(self, n_keep: int):
        """
        Args:
            n_keep (int): The number of estimators to keep

        Raises:
            ValueError: If the number of estimators to keep is less than or equal to 0
        """
        if n_keep <= 0:
            raise ValueError(f"Argument n_keep must be greater than 0 not {n_keep}.")
        self._n_keep = n_keep

        # remove estimators beyond n_keep
        if n_keep < len(self._estimators):
            difference = len(self._estimators) - n_keep
            self._estimators = deque(list(self._estimators)[difference:])

    def fit_update(
        self, x: torch.Tensor, t: torch.Tensor, limit: torch.LongTensor = None
    ):

        self._base_estimator.fit(x, t, limit)
        if len(self._estimators) == self._n_keep:
            self._estimators.rotate(-1)
            self._estimators[-1] = deepcopy(self._base_estimator)
        else:
            self._estimators.append(deepcopy(self._base_estimator))

        self._fitted = True

    def forward(self, x: torch.Tensor) -> torch.Tensor:

        if self._fitted is False:
            return self._base_estimator(x)
        # TODO: This can be optimized so that it does not convert to a
        # tensor.
        votes = torch.stack([estimator(x) for estimator in self._estimators])
        return self._voter(votes)

    @property
    def n_estimators(self) -> estimators.ScikitEstimator:
        return len(self._estimators)

    @property
    def fitted(self) -> bool:
        return self._fitted


class VoterEnsembleMachine(LearningMachine, FeatureIdxStepX, FeatureIdxStepTheta):
    """Machine that runs an ensemble of sub machines"""

    def __init__(
        self,
        base_estimator: estimators.ScikitEstimator,
        n_keep: int,
        step_x: FeatureIdxStepX,
        loss: Loss,
        preprocessor: nn.Module = None,
    ):
        """

        Args:
            base_estimator (scikit.ScikitEstimator): Base estimator
            n_keep (int): Number of estimators to keep each round
            step_x (StepX): StepX to update machine with
            loss (Loss): The loss to evaluate the machine with
            preprocessor (nn.Module, optional): Module to execute before . Defaults to None.
        """
        super().__init__()
        self._module = VoterEnsemble(base_estimator, n_keep)
        self._loss = loss
        self._step_x = step_x
        self._preprocessor = preprocessor

    def assess_y(self, y: IO, t: IO, reduction_override: str = None) -> AssessmentDict:
        return self._loss.assess_dict(y[0], t[0], reduction_override)

    def step(
        self, conn: Conn, state: State, from_: IO = None, feature_idx: Idx = None
    ) -> Conn:
        """Update the machine

        Args:
            conn (Conn): the connection to train on
            state (State): State for training
            from_ (IO, optional): the input to the previous machine. Defaults to None.
            feature_idx (Idx, optional): A limit on the connections that get trained. Defaults to None.

        Returns:
            Conn:
        """
        x = conn.step.x[0]
        if self._preprocessor is not None:
            x = self._preprocessor(x)

        t = conn.step.t[0]
        self._module.fit_update(
            x, t, feature_idx.tolist() if feature_idx is not None else None
        )

        return conn.connect_in(from_)

    def step_x(self, conn: Conn, state: State, feature_idx: Idx = None) -> Conn:
        """Update the input

        Args:
            conn (Conn):  the connection to do step_x on
            state (State): State for training
            feature_idx (Idx, optional): A limit on the connections that get trained. Defaults to None.

        Returns:
            Conn: connection
        """
        return self._step_x.step_x(conn, state, feature_idx)

    def forward(self, x: IO, state: State, detach: bool = True) -> IO:
        """

        Args:
            x (IO): send the input through the voting ensemble
            state (State): the state for learning
            detach (bool, optional): whether to detach the output or not. Defaults to True.

        Returns:
            IO: _description_
        """

        x = x[0]
        if self._preprocessor is not None:
            x = self._preprocessor(x)

        y = IO(self._module(x))
        return y.out(detach=detach)

    @property
    def fitted(self) -> bool:
        """
        Returns:
            bool: whether the module was fitted or not
        """
        return self._module.fitted



"""
Core Modules for Zen

Modules:
LearningMachine - nn.Module which has the ability to learn on its own


Optimization:
The following classes can be used to add flexibility to optimziation
StepX - Optimizer for updating the inputs of a learning machine
StepTheta - Optimizer for updating the parameters of a learning machine

Connections:
Conn - Base class for all connections
Node - Class used for learning machine connections

Other
Loop - Loop over data
"""

import typing

# 1st party
from abc import ABC, abstractmethod
from collections import deque

# 3rd party
import torch
import torch.nn as nn

from .. import utils as base_utils

# local
from .assess import AssessmentDict, Loss, ThLoss
from .component import Learner
from .state import IDable, State


class IO(object):
    """Handles IO into and out of learning machines
    to give a consistent system to handle it
    """

    def __init__(self, *x, detach: bool = False, names: typing.List[str] = None):
        """initializer

        Args:
            detach (bool, optional): The values making up the IO. Defaults to False.
            names (typing.List[str], optional): The name of each value. Defaults to None.
        """
        super().__init__()

        self._x = []
        self._freshened = False
        self._singular = len(x) == 1
        for x_i in x:
            if isinstance(x_i, torch.Tensor) and detach:
                x_i = x_i.detach()

            self._x.append(x_i)
        self._names = enumerate(dict(names or []))

    def freshen(self, inplace: bool = False) -> bool:
        """Set the values of the IO

        Args:
            inplace (bool, optional): _description_. Defaults to False.

        Returns:
            bool: _description_
        """
        if self._freshened:
            return False

        self._x = [base_utils.freshen(x_i, inplace=inplace) for x_i in self._x]
        self._freshened = True
        return self

    def items(self) -> typing.Dict:
        return dict(enumerate(self._x))

    @property
    def vals(self) -> typing.List:
        """the values making up the IO

        Returns:
            list: The values in the IO
        """
        return [*self._x]

    def __getitem__(self, idx: int):
        """Retrieve item from the IO

        Args:
            idx (int): The index to retrieve for

        Returns:
            the value at the index
        """
        return self._x[idx]

    def to(self, device) -> "IO":
        """Change the device of all tensors in the IO

        Args:
            device: The device to change the convert to

        Returns:
            IO: self
        """

        if device is None:
            return self

        self._x = [
            x_i.to(device) if isinstance(x_i, torch.Tensor) else x_i for x_i in self._x
        ]
        if self._freshened:
            self._freshened = False
            self.freshen(self._x)
        return self

    @property
    def names(self) -> typing.List[str]:
        """
        Returns:
            typing.List[str]: The names of all of the fields
        """
        return [*self._names]

    def __len__(self) -> int:
        """
        Returns:
            int: The number of fields in the IO
        """
        return len(self._x)

    def __iter__(self) -> typing.Iterator:
        """
        Returns:
            typing.Iterator: Iterator of all of the elements
        """
        return iter(self._x)

    def clone(self, detach: bool = True) -> "IO":
        """create a copy of the of all of the tensors

        Args:
            detach (bool, optional): Whether to clone the gradients. Defaults to True.

        Returns:
            IO: The cloned IO
        """
        x = [torch.clone(x_i) for x_i in self._x]
        result = IO(*x, detach=detach, names=self._names)
        if not detach:
            result._freshened = self._freshened
        return result

    def detach(self) -> "IO":

        return IO(*self._x, detach=True, names=self._names)

    def out(self, detach: bool = True, clone: bool = False) -> "IO":
        y = self
        if detach:
            y = y.detach()
        if clone:
            y = y.clone()
        return y

    def is_empty(self) -> bool:
        return len(self) == 0


class Idx(object):
    def __init__(self, idx=None, dim: int = 0):
        """initializer

        Set an index on the IO to

        usage: Use when the connection should retrieve a subset of the values
        in the IO

        Args:
            idx (optional): The values to index by. Defaults to None.
        """
        if not isinstance(idx, torch.LongTensor) and idx is not None:
            idx = torch.LongTensor(idx)
        self.dim = dim
        self.idx = idx

    def idx_th(
        self, *x: torch.Tensor
    ) -> typing.Union[typing.Tuple[torch.Tensor], torch.Tensor]:

        if self.idx is not None:
            x = [x_i.index_select(self.dim, self.idx) for x_i in x]

        return x

    def tolist(self) -> typing.Union[None, typing.List[int]]:
        if self.idx is None:
            return None
        return self.idx.tolist()

    def idx_list(self) -> typing.List[int]:
        """

        Returns:
            typing.List[int]: _description_
        """

        result = []
        for i in self.idx:
            result.append(self.idx[i.item()])
        return result

    def __call__(self, x: IO, detach: bool = False) -> IO:

        selected = self.idx_th(*x.vals)

        result = IO(*selected, detach=detach, names=x.names)
        if x._freshened and not detach:
            result._freshened = True
        return result

    def update(self, source: IO, destination: IO):
        for source_i, destination_i in zip(source, destination):
            if destination_i.requires_grad:
                requires_grad = True
                destination_i.detach_().requires_grad_(False)
            else:
                requires_grad = False
            if self.idx is not None:
                destination_i[self.idx] = source_i
            else:
                destination_i[:] = source_i
            if requires_grad:
                destination_i.requires_grad_(True).retain_grad()

    def update_th(self, source: torch.Tensor, destination: torch.Tensor):
        if self.idx is not None:
            destination[self.idx] = source
        else:
            destination[:] = source

    def sub(self, idx: "Idx") -> "Idx":
        if not isinstance(idx, Idx):
            idx = Idx(idx)

        if idx.idx is None:
            return self
        elif self.idx is None:
            return idx

        return Idx(self.idx[idx.idx])

    def __len__(self) -> int:
        return len(self.idx)


class LayerIO(object):
    """IO wrapper for the inputs, outputs and targets for a layer"""

    @staticmethod
    def to_io(io: IO = None):
        if io is None:
            return IO()
        elif not isinstance(io, IO):
            return IO(io)
        return io

    # TODO: better handle the case when None
    def __init__(self, x: IO = None, t: IO = None, y: IO = None):  # , idx: Idx=None):
        """initializer

        Args:
            x (IO, optional): IO for the input. Defaults to None.
            t (IO, optional): IO for the target. Defaults to None.
            y (IO, optional): IO for the output. Defaults to None.
            idx (Idx, optional): Index to index the IO by. Defaults to None.
        """
        self._x = self.to_io(x)
        self._t = self.to_io(t)
        self._y = self.to_io(y)

    @property
    def x(self) -> "IO":
        """
        Returns:
            IO: the IO for x. If an index is specified it will
            be indexed
        """
        return self._x
        # return self._x.get(self._idx)

    @x.setter
    def x(self, x: IO) -> "IO":
        """Sets the value for x. Cannot use if is_indexed is True

        Args:
            x (IO): The IO to set x to

        Raises:
            ValueError: If is_index is true

        Returns:
            IO: The updated IO
        """
        # TODO: Not sure if i want to leave this
        # if self.idx is not None:
        #    raise ValueError("Cannot set x if idx is set")
        self._x = x

    @property
    def t(self) -> "IO":
        return self._t

    @t.setter
    def t(self, t: IO) -> "IO":
        # TODO: Not sure if i want to leave this
        # if self.idx is not None:
        #    raise ValueError("Cannot set t if idx is set")
        self._t = t

    @property
    def y(self) -> "IO":

        return self._y  # .get(self._idx)

    @y.setter
    def y(self, y: IO) -> "IO":
        self._y = y

    def __iter__(self) -> typing.Iterator[IO]:
        yield self.x
        yield self.t
        yield self.y

    # TODO: Review if needed
    def out(self, detach: bool = True) -> "LayerIO":
        x = self._x.detach() if self.x is not None and detach else self._x
        t = self._t.detach() if self.t is not None and detach else self._t
        y = self._y.detach() if self.y is not None and detach else self._y
        return LayerIO(
            x,
            t,
            y,  # idx=self._idx if use_idx else None
        )


class Conn(object):
    """Class to connect two learning machines"""

    def __init__(
        self,
        out_x: IO,
        out_t: IO,
        inp_x: IO = None,
        inp_t: IO = None,
        out_y: IO = None,
        inp_y: IO = None,
        state=None,
    ):
        super().__init__()
        inp_t = inp_t or out_x
        self._inp = LayerIO(inp_x, inp_t, inp_y)
        self._out = LayerIO(out_x, out_t, out_y)

        self.state = state or State()
        self.tie_step = self.tie_inp
        self.tie_step_x = self.tie_out

    def tie_inp(self, detach: bool = False) -> "Conn":
        """Set the target of the incoming layer to the same as the input of the outgoing layer

        Args:
            detach (bool, optional): Whether to detach after setting. Defaults to False.

        Returns:
            Conn: _description_
        """
        self._inp.t = self._out.x
        if detach:
            self._inp.t.detach()
        return self

    def tie_out(self, detach: bool = False) -> "Conn":
        """Set the input of the outgoing layer to the same as the output of the outgoing layer

        Args:
            detach (bool, optional): Whether to detach after setting. Defaults to False.

        Returns:
            Conn: self
        """
        self._out.x = self._inp.y
        if detach:
            self._out.x.detach()
        return self

    @property
    def inp(self) -> LayerIO:
        """
        Returns:
            LayerIO: The indexed LayerIO for the incoming layer if index is set else the base
        """
        return self._inp

    @property
    def out(self) -> LayerIO:
        """
        Returns:
            LayerIO: The indexed LayerIO for the outgoing layer if index is set else the base
        """
        return self._out

    @property
    def step_x(self) -> LayerIO:
        """
        Returns:
            LayerIO: The indexed LayerIO for the outgoing layer if index is set else the base
        """
        return self.out

    @property
    def step(self) -> LayerIO:
        """
        Returns:
            LayerIO: The indexed LayerIO for the incoming layer if index is set else the base
        """
        return self.inp

    def connect_in(self, from_in_x: IO = None) -> "Conn":
        """Connect to a new incoming layer. The current incoming layer becomes the outgoing layer

        Args:
            from_in_x (IO, optional): The x value for the incoming layer. Defaults to None.
            use_idx (bool, optional): Whether the idx should be passed on to the new connection. Defaults to True.

        Returns:
            Conn: The connection with the incoming layer
        """
        return Conn(self._inp.x, self._inp.t, inp_x=from_in_x)

    def retie(self, inp_x: IO) -> "IO":
        """

        Args:
            inp_x (IO, optional): The new x for the incoming layer

        Returns:
            IO: the previous x value for the incoming layer
        """
        old_x = self.inp.x
        self._inp.x = inp_x
        return old_x

    @classmethod
    def from_layer_io(self, incoming: LayerIO, outgoing: LayerIO) -> "Conn":

        return Conn(
            outgoing.x, outgoing.t, incoming.x, incoming.t, outgoing.y, incoming.y
        )


def idx_io(io: IO, idx: Idx = None, detach: bool = False) -> IO:

    if idx is not None:
        io = idx(io)

    return io.out(detach)


def idx_layer_io(layer_io: LayerIO, idx: Idx = None, detach: bool = False) -> LayerIO:

    return LayerIO(
        idx_io(layer_io.x, idx, detach),
        idx_io(layer_io.t, idx, detach),
        idx_io(layer_io.y, idx, detach),
    )


def idx_conn(conn: Conn, idx: Idx = None, detach: bool = False) -> Conn:

    return Conn.from_layer_io(
        idx_layer_io(conn.step, idx, detach),
        idx_layer_io(conn.step_x, idx, detach),
    )


def update_step_x(
    source: Conn,
    destination: Conn,
    idx: Idx = None,
    tie_step: bool = True,
    detach: bool = True,
) -> Conn:

    update_io(source.step_x.x, destination.step_x.x, idx)
    if tie_step:
        destination.tie_step(detach)
    return destination


def idx_th(x: torch.Tensor, idx: Idx = None, detach: bool = False) -> torch.Tensor:

    if idx is not None:
        x = idx.idx_th(x)

    if detach:
        x = x.detach()
    return x


# TODO: DEBUG. This is not working for some reason
def update_io(source: IO, destination: IO, idx: Idx = None) -> IO:

    if idx is None:
        idx = Idx()
    idx.update(source, destination)
    return destination


def update_tensor(
    source: torch.Tensor, destination: torch.Tensor, idx: Idx = None
) -> torch.Tensor:

    if idx is None:
        idx = Idx()
    idx.update_th(source, destination)
    return destination


class T(Conn):
    def __init__(self, t: IO, inp_x: IO = None):
        """Create a 'Target' connection for the global target

        Args:
            t (IO): the target to set
            inp_x (IO, optional): the x for the incoming layer. Defaults to None.
        """
        super().__init__(t, t, inp_x=inp_x)


class StepXHook(ABC):
    @abstractmethod
    def __call__(self, conn: Conn, state: State, **kwargs) -> Conn:
        pass


class StepHook(ABC):
    @abstractmethod
    def __call__(self, conn: Conn, state: State, **kwargs) -> Conn:
        pass


class StepX(ABC):
    """Base class for updating the input (target)
    Use to decouple the optimization of the input from the learning
    machine definition
    """

    def __init__(self):
        self._step_x_hook_initialized = True
        self._step_x_prehooks = []
        self._step_x_posthooks = []
        self._base_step_x = self.step_x
        self.step_x = self._step_x_hook_runner

    @abstractmethod
    def step_x(self, conn: Conn, state: State) -> Conn:
        pass

    def __call__(self, conn: Conn, state: State, *args, **kwargs) -> Conn:
        return self.step_x(conn, state, *args, **kwargs)

    def _step_x_hook_runner(self, conn: Conn, state: State, *args, **kwargs) -> Conn:
        """Call step x wrapped with the hooks

        Args:
            conn (Conn): The connection to run on
            state (State): The current state

        Returns:
            Conn: The connection
        """

        for prehook in self._step_x_prehooks:
            conn = prehook(conn, state)

        conn = self._base_step_x(conn, state, *args, **kwargs)

        for posthook in self._step_x_posthooks:
            conn = posthook(conn, state)

        return conn

    def step_x_prehook(self, hook: StepXHook):
        """Add hook to call before StepX

        Args:
            hook (StepXHook): The hook to add
        """
        if not hasattr(self, "_step_x_hook_initialized"):
            self.__init__()
        self._step_x_prehooks.append(hook)

    def step_x_posthook(self, hook: StepXHook):
        """Add hook to call after StepX

        Args:
            hook (StepXHook): The hook to add
        """
        if not hasattr(self, "_step_x_hook_initialized"):
            self.__init__()
        self._step_x_posthooks.append(hook)


class StepTheta(ABC):
    """Base class for updating the parameters
    Use to decouple the optimization of the parameters from the core
    machine definition
    """

    def __init__(self):

        self._step_hook_initialized = True
        self._step_prehooks = []
        self._step_posthooks = []
        self._base_step = self.step
        self.step = self._step_hook_runner

    @abstractmethod
    def step(self, conn: Conn, state: State, from_: IO = None) -> Conn:
        pass

    def __call__(
        self, conn: Conn, state: State, from_: IO = None, *args, **kwargs
    ) -> Conn:
        return self.step(conn, state, from_, *args, **kwargs)

    def _step_hook_runner(
        self, conn: Conn, state: State, from_: IO = None, *args, **kwargs
    ) -> Conn:
        """Call step wrapped with the hooks

        Args:
            conn (Conn): The connection to run on
            state (State): The current state

        Returns:
            Conn: The connection
        """
        for prehook in self._step_prehooks:
            conn = prehook(conn, state)

        conn = self._base_step(conn, state, from_, *args, **kwargs)

        for posthook in self._step_posthooks:
            conn = posthook(conn, state)

        return conn

    def step_prehook(self, hook: StepHook):
        """Add hook to call before StepTheta

        Args:
            hook (StepHook): The hook to add
        """
        if not hasattr(self, "_step_hook_initialized"):
            self.__init__()
        self._step_prehooks.append(hook)

    def step_posthook(self, hook: StepHook):
        """Add hook to call after StepTheta

        Args:
            hook (StepHook): The hook to add
        """
        if not hasattr(self, "_step_hook_initialized"):
            self.__init__()
        self._step_posthooks.append(hook)


class BatchIdxStepTheta(StepTheta):
    @abstractmethod
    def step(
        self, conn: Conn, state: State, from_: IO = None, batch_idx: Idx = None
    ) -> Conn:
        pass


class FeatureIdxStepTheta(StepTheta):
    @abstractmethod
    def step(
        self, conn: Conn, state: State, from_: IO = None, feature_idx: Idx = None
    ) -> Conn:
        pass


class BatchIdxStepX(StepX):
    @abstractmethod
    def step_x(self, conn: Conn, state: State, batch_idx: Idx = None) -> Conn:
        pass


class FeatureIdxStepX(StepX):
    @abstractmethod
    def step_x(self, conn: Conn, state: State, feature_idx: Idx = None) -> Conn:
        pass


class LearningMachine(nn.Module, Learner, StepTheta, StepX, IDable, ABC):
    def device(self) -> torch.device:
        """Convenience method to get the device for the machine
        Chooses the first parameter. Assumes all sub machines have the same device

        Returns:
            torch.device: Device of the learning machine
        """

        try:
            p = next(self.parameters())
            return p.device
        except StopIteration:
            return None

    def to_my_device(
        self, *io: IO
    ) -> typing.Union[typing.Tuple[torch.device], torch.device]:
        """Convenience method to convert x to the device of the machine.
        Assumes that the device will be the same
        """
        device = self.device()
        if device is None:
            return io if len(io) > 1 else io[0]
        if len(io) == 1:
            return io[0].to(device)
        return tuple(io_i.to(device) for io_i in io)

    @property
    def id(self) -> str:
        return str(id(self))

    @abstractmethod
    def assess_y(self, y: IO, t: IO, reduction_override: str = None) -> AssessmentDict:
        """Assess the learning machine

        Args:
            y (IO): the output of the machine
            t (IO): the target
            reduction_override (str, optional): Value to override
              the reduction by. If None will not override. Defaults to None.

        Returns:
            AssessmentDict: The assessment of the machine
        """
        pass

    def assess(
        self,
        x: IO,
        t: IO,
        reduction_override: str = None,
        state: State = None,
        detach: bool = False,
    ) -> AssessmentDict:
        """Assess the learning machine

        Args:
            x (IO): _description_
            t (IO): _description_
            reduction_override (str, optional): Value to override the
              reduction by. If None will not override. Defaults to None.
            state (State, optional): Defaults to None.
            detach (bool, optional): Whether to detach the output.
              Defaults to False.

        Returns:
            AssessmentDict: The assessment of the machine
        """
        y = self(x, state=state, detach=detach)
        return self.assess_y(y, t, reduction_override=reduction_override)

    @abstractmethod
    def forward(self, x: IO, state: State, detach: bool = True) -> IO:
        """
        Args:
            x (IO): The input to the machine
            state (State)
            detach (bool, optional): Whether to detach the output. Defaults to True.

        Returns:
            IO: The output fo the machine
        """
        raise NotImplementedError

    def __call__(self, x: IO, state: State = None, detach: bool = True) -> IO:
        """
        Args:
            x (IO): The input to the machine
            state (State, optional): Defaults to None.
            detach (bool, optional): Whether to detach the output. Defaults to True.

        Returns:
            IO: The output fo the machine
        """
        return super().__call__(x, state or State(), detach)

    def learn(
        self,
        x: IO,
        t: IO,
        state: State = None,
        return_x_step: bool = False,
        clear_state: bool = False,
        reduction_override: str = None,
    ) -> AssessmentDict:
        """Learn method . This includes cleanup and initialization so it is easier to use in practice
        than step

        Args:
            x: The input to the machine
            t: The target to the machine
            state (State, optional): The current learning state. Defaults to None.
            return_step (bool, optional): Whether to return step_x based on the inputs. Defaults to False.
            clear_state (bool, optional): Whether to clear teh state for the machine. Defaults to False.

        Returns:
            AssessmentDict: _description_
        """
        if not self.training:
            self.train()
        x, t = self.to_my_device(x, t)
        state = State()
        y = self(x, state)
        assessment = self.assess_y(y, t, reduction_override=reduction_override)

        conn = self.step(T(t, inp_x=x), state)
        if clear_state:
            state.clear(self)
        if return_x_step:
            return conn, assessment
        return assessment

    def full_step(
        self, conn: "Conn", state: State, from_: IO = None, clear_state: bool = False
    ) -> Conn:
        conn = self.step(conn, state, from_=from_)
        conn = self.step_x(conn, state)

        if clear_state:
            state.clear(self)
        return conn

    def test(self, x: IO, t: IO) -> AssessmentDict:
        """Assess the machine in "testing" mode

        Args:
            x (IO): the input to the machine
            t (IO): the output to the machine

        Returns:
            AssessmentDict: The assessment
        """
        if self.training:
            self.eval()
        with torch.no_grad():
            x, t = self.to_my_device(x, t)
            return self.assess_y(self(x), t).cpu().detach()


class EmissionStack(object):
    def __init__(self, *emissions: IO):
        """Convenience wrapper for deque to simplify recording emissions for the step method

        usage:
        def forward(self, x) -> IO:
            ...
            emissions = EmissionStack()
            x = emissions(layer(x))
            state[self, 'emissions'] = emissions
            ...

        def step(self, ...):
            ...
            layer.step(conn, state, from_=state[self, 'emissions'].pop())

        """
        self._stack = deque(emissions)

    def __call__(self, io: IO) -> IO:
        """Add an element to the stack

        Args:
            io (IO): Element to add

        Returns:
            IO: the element that was added
        """

        self._stack.append(io)
        return io

    def __len__(self) -> int:
        return len(self._stack)

    def stack_on(self, io: IO):
        """Restack the stack by placing it on another vlaue

        Args:
            io (IO): the io to stack the current stack onto
        """

        self._stack.insert(0, io)

    def pop(self) -> typing.Union[IO, None]:
        """Pop off the last element in the stack. Returns None if empty

        Raises:
            IndexError: If there are no elements left in the stack

        Returns:
            IO: the last element
        """

        try:
            return self._stack.pop()
        except IndexError:
            return None
            # raise IndexError("No more elements left in the EmissionStack to pop")

    def __iter__(self):
        """
        LIFO Iteration over the stack
        """

        for io in reversed(self._stack):
            yield io


"""

m1.step( clear=True)


zen.repeat(machine1, machine2, 3)
   .step(machine2, machine3)
   .sequence()
   .stepper()
   .backward()


stepper.backward()

zen.step_pipeline(
  zen.RepeatStep(machine1, out=machine2),
  zen.Step(machine2, out=machine3),
  zen.SequenceStep(machine3, machine4, machine5)
)

"""


class NullLearner(LearningMachine):
    def __init__(self, loss: Loss = None):
        """Machine that does not actually learn.

        usage: Use when an intermediary layer should not perform any operation on the backward
        pass. Can use

        Args:
            loss (Loss, optional): The loss to evaluate by. Defaults to None.
        """
        super().__init__()
        self.loss = loss or ThLoss(nn.MSELoss, reduction="none")
        # self.step_x_learner = step_x_learner

    def assess_y(self, y: IO, t: IO, reduction_override: str = None) -> AssessmentDict:
        return self.loss.assess_dict(*y, *t, reduction_override)

    def step(self, conn: Conn, state: State, from_: IO = None) -> Conn:
        return Conn(conn.out.x, conn.out.t, from_, conn.out.t)

    def step_x(self, conn: Conn, state: State) -> Conn:
        return conn

    def forward(self, x: IO, state: State):
        return x


# def forward(layer: nn.Module, x: IO, freshen: bool=True, multi_out: bool=False, detach: bool=True) -> IO:
#     """
#     Args:
#         layer (nn.Module): layer to pass the io through
#         x (IO): the input into the layer
#         freshen (bool, optional): Whether or not to freshen the input. Defaults to False.
#         multi_out (bool, optional): _description_. Defaults to False.

#     Returns:
#         IO: returned IO
#     """

#     if freshen:
#         x.freshen(False)
#     y = layer(*x)
#     if multi_out:
#         y = IO(*y)
#     else:
#         y = IO(y)

#     return y.out(detach)



# 1st party
from abc import ABC, abstractmethod

# 3rd party
import torch

from ..utils import to_signed_neg, to_zero_neg
from .core import Individual, Population
from .exploration import EqualsAssessmentDist

# local
from .selectors import BinaryProbSelector, SlopeSelector


class PopulationModifier(ABC):
    """Modifies an Individual based on the Population"""

    @abstractmethod
    def __call__(self, original: Individual, population: Population) -> Individual:
        pass

    @abstractmethod
    def spawn(self) -> "PopulationModifier":
        pass


class SelectionModifier(ABC):
    """Modifies an Individual with the result of a selection (an Individual)"""

    @abstractmethod
    def __call__(self, original: Individual, population: Individual) -> Individual:
        pass

    @abstractmethod
    def spawn(self) -> "SelectionModifier":
        pass


class SlopeModifier(PopulationModifier):
    """Modifies an individual based on the slope of a population"""

    def __init__(
        self, momentum: float, lr: float = 0.1, x: str = "x", maximize: bool = False
    ):
        """initializer

        Args:
            momentum (float): The amount of momentum for the slope
            lr (float, optional): The "learning" rate (i.e. how much to go in the direction).
                Defaults to 0.1.
            x (str, optional): The name for the x value. Defaults to "x".
            maximize (bool, optional): Whether to maximize or minimize.
              If minimizing, the sign of the lr will be reversed. Defaults to False.
        """

        self._slope_selector = SlopeSelector(momentum)
        self._lr = lr if maximize else -lr
        self.x = x
        self._momentum = momentum
        self._multiplier = 1 if maximize else -1

    @property
    def lr(self) -> float:
        return self._lr

    @lr.setter
    def lr(self, lr):
        self._lr = abs(lr) * self._multiplier

    @property
    def maximize(self) -> bool:
        return self._multiplier == 1

    @maximize.setter
    def maximize(self, maximize):
        self._multiplier = 1 if maximize else -1

    def __call__(self, original: Individual, population: Population) -> Individual:
        x = original[self.x]
        slope = self._slope_selector(population)[self.x]
        return Individual(**{self.x: x + self.lr * slope})

    def spawn(self) -> "SlopeModifier":
        return SlopeModifier(self._momentum, self.lr, self.x, self._multiplier == 1)


class BinaryAdjGaussianModifier(PopulationModifier):
    """Choose whether each value should be positive or negative based on the value of each"""

    def __init__(self, k: int, x: str = "x", zero_neg: bool = False):
        """initializer
        Args:
            k (int): The number of
            x (str, optional): The name of the x value. The individual/population must have it. Defaults to 'x'.
            zero_neg (bool, optional): _description_. Defaults to False.

        Raises:
            ValueError: _description_
        """
        # k is an upper bound. If the dimension size is smaller
        # the dimension size will be used
        super().__init__()
        if k <= 0:
            raise ValueError(f"Argument k must be greater than 0 not {k}")
        self.x = x
        self.k = k
        self.zero_neg = zero_neg
        self.pos_assessment_calc = EqualsAssessmentDist(1)
        self.neg_assessment_calc = EqualsAssessmentDist(-1 if not zero_neg else 0)

    def __call__(self, original: Individual, population: Population) -> Individual:
        """Estimate the value of positives and negatives and update

        Args:
            original (Individual): The individual to modify
            population (Population): The population to modify based on

        Returns:
            Individual: The modified individual
        """

        assessments = population.stack_assessments()
        x_pop = population[self.x]
        x_ind = original[self.x]
        k = min(self.k, x_ind.shape[1])

        # calculate the average value of positives and negatives
        pos = self.pos_assessment_calc.mean(assessments, x_pop)
        neg = self.neg_assessment_calc.mean(assessments, x_pop)

        difference = pos - neg
        difference[pos.isnan()] = -1
        difference[neg.isnan()] = 1
        base_result = torch.sign(difference)
        # make it into a "maximization" problem
        adj_difference = 1 / (difference.abs() + 1.0)
        # reduce to 0 if no change
        adj_difference[x_ind == base_result] = 0.0
        _, best_indices = adj_difference.topk(dim=1, k=k, largest=True)
        best_differences = difference.gather(index=best_indices, dim=1)
        result = torch.sign(best_differences)
        if self.zero_neg:
            result = to_zero_neg(result)
        x_ind = x_ind.scatter(1, best_indices, result)

        return Individual(**{self.x: x_ind})

    def spawn(self) -> "BinaryAdjGaussianModifier":
        return BinaryAdjGaussianModifier(self.k, self.x, self.zero_neg)


class BinaryGaussianModifier(PopulationModifier):
    """Choose up to k updates for the individual. If there are no
    updates that produce a large change in value. None will be updated"""

    def __init__(self, k: int, x: str = "x", zero_neg: bool = False):
        super().__init__()
        if k <= 0:
            raise ValueError(f"Argument k must be greater than 0 not {k}")
        self.x = x
        self.k = k
        self.zero_neg = zero_neg
        self.pos_assessment_calc = EqualsAssessmentDist(1)
        self.neg_assessment_calc = EqualsAssessmentDist(-1 if not zero_neg else 0)

    def __call__(self, original: Individual, population: Population) -> Individual:

        assessments = population.stack_assessments()
        x_pop = population[self.x]
        x_ind = original[self.x]

        k = min(self.k, x_ind.shape[1])
        pos = self.pos_assessment_calc.mean(assessments, x_pop)
        neg = self.neg_assessment_calc.mean(assessments, x_pop)

        difference = pos - neg
        base_result = torch.sign(difference)
        _, best_indices = difference.abs().topk(dim=1, k=k, largest=False)
        best_results = base_result.gather(dim=1, index=best_indices)
        result = x_ind.scatter(1, best_indices, best_results)

        if self.zero_neg:
            result = to_zero_neg(result)
        return Individual(**{self.x: result})

    def spawn(self) -> "BinaryGaussianModifier":
        return BinaryGaussianModifier(self.k, self.x, self.zero_neg)


class BinaryProbModifier(PopulationModifier):
    def __init__(self, zero_neg: bool, x: str = "x") -> None:
        """initializer

        Args:
            zero_neg (bool): Whether the negative value is represented by zero
            x (str, optional): The name of the value to modify. Defaults to "x".
        """
        super().__init__()
        self.zero_neg = zero_neg
        self._binary_selector = BinaryProbSelector()
        self.x = x

    def __call__(self, original: Individual, population: Population) -> Individual:
        """

        Args:
            original (Individual): The individual to modify
            population (Population): The population to modify based on

        Returns:
            Individual
        """

        updated = self._binary_selector(population)
        original = original[self.x]
        neg_count = updated[self._binary_selector.neg_count]
        pos_count = updated[self._binary_selector.pos_count]
        x = updated[self.x]
        x = (torch.rand(x.size(), device=x.device) < x).type_as(x)

        no_change = (pos_count == 0) | (neg_count == 0)

        if not self.zero_neg:
            original = (original + 1) / 2
        result = (~no_change).float() * x + (no_change).float() * original
        if not self.zero_neg:
            result = to_signed_neg(result)
        return Individual(x=result)

    def spawn(self) -> "BinaryProbModifier":
        return BinaryProbModifier(self.zero_neg, self.x)


class KeepModifier(SelectionModifier):
    """Modify the original based on the selection by keeping the values in
    the individual with a set probability"""

    def __init__(self, keep_p: float):
        if not (0 < keep_p < 1.0):
            raise ValueError(f"{keep_p} must be between 0 and 1.")
        self.keep_p = keep_p

    def __call__(self, original: Individual, selection: Individual) -> Individual:
        """Randomly choose whether to select original or selection for each value

        Args:
            original (Individual): The individual to modify
            population (Population): The population to modify based on

        Returns:
            Individual: The modified individual
        """
        new_values = {}
        for k, v in original:
            if k in selection:
                keep = (torch.rand_like(v) < self.keep_p).type_as(v)
                new_values[k] = keep * v + (1 - keep) * selection[k]

        return Individual(**{new_values})

    def spawn(self) -> "KeepModifier":
        return KeepModifier(self.keep_p)
